/************************************************************************/
/* Copyright (C) 2008-2009 Markus Strobl <markus.strobl@gmx.at>         */
/*                                                                      */
/* cpupowerd is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by    */
/* the Free Software Foundation; either version 2 of the License.       */
/*                                                                      */
/* cpupowerd is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of           */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */
/*                                                                      */
/* You should have received a copy of the GNU General Public License    */
/* along with GNU Emacs; see the file COPYING.  If not, write to        */
/* the Free Software Foundation, Inc., 59 Temple Place - Suite 330,     */
/* Boston, MA 02111-1307, USA.                                          */
/************************************************************************/
/* $Id: cpu.c 369 2009-01-01 13:05:07Z markus $                         */
/************************************************************************/


/* Includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cpu.h"
#include "cpupowerd.h"
#include "libcpupowerd.h"
#ifdef __linux__
#include "cpu_linux.h"
#elif __FreeBSD__
#include "cpu_bsd.h"
#endif


/* Function pointers */
int (*ptr_cpu_get_load) (int, unsigned long long *, unsigned long long *, float *);		/* cpu_get_load */
int (*ptr_cpu_get_laptop_mode) (int *); 												/* cpu_get_laptop_mode */
int (*ptr_cpu_get_mastercpuids) (int *, int **, int *); 								/* cpu_get_mastercpuids */
int (*ptr_cpu_get_affectedcpuids) (int, int **, int *);									/* cpu_get_affected_cpudis */
int (*ptr_cpu_get_frequencies) (int, int **, int *);									/* cpu_get_frequencies */
int (*ptr_cpu_get_frequency) (int, int *);												/* cpu_get_frequency */
int (*ptr_cpu_set_frequency) (int, int);												/* cpu_set_frequency */
int (*ptr_cpu_get_governor) (int, char *, int);											/* cpu_get_governor */
int (*ptr_cpu_set_governor) (int, char *);												/* cpu_set_governor */
int (*ptr_cpu_tini) (void);																/* cpu_tini */


/* Load for mastercpu */
typedef struct load
{
	int mastercpuid;							/* mastercpuid */
	int cpuid;									/* cpuid */
	unsigned long long usage;					/* usage cpu */
	unsigned long long total;					/* total cpu */
} load_t;


/* Global variables */
static load_t **loads; 							/* Global array that contains a pointer to all cpu loads. */


/* Functions */
/* Init */
int cpu_init(int *cpuids)
{
	int i, j;
	int result;
	int counter;
	int cpuids_quantity;
	int mastercpuids_quantity;
	int affectedcpuids_quantity;
	float load_tmp;
	int *affectedcpuids;
	int *mastercpuids;
	load_t *load;

	#ifdef __linux__
	result = cpu_linux_init(&ptr_cpu_get_load, &ptr_cpu_get_laptop_mode,
						    &ptr_cpu_get_mastercpuids, &ptr_cpu_get_affectedcpuids,
							&ptr_cpu_get_frequencies, &ptr_cpu_get_frequency,
							&ptr_cpu_set_frequency, &ptr_cpu_get_governor,
							&ptr_cpu_set_governor, &ptr_cpu_tini);
	#elif __FreeBSD__
	result = cpu_bsd_init(&ptr_cpu_get_load, &ptr_cpu_get_laptop_mode,
					 	  &ptr_cpu_get_mastercpuids, &ptr_cpu_get_affectedcpuids,
						  &ptr_cpu_get_frequencies, &ptr_cpu_get_frequency,
						  &ptr_cpu_set_frequency, &ptr_cpu_get_governor,
						  &ptr_cpu_set_governor, &ptr_cpu_tini);
	#else
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Your operating system isn't yet supported!\n");
		result = NOTOK;
	#endif

	if (result != OK)
		return NOTOK;

	cpuids_quantity = 0;
	while (cpuids[cpuids_quantity] >= 0)
		cpuids_quantity++;

	/* Allocate memory for load pointer array */
	if ((loads = (load_t **)malloc(sizeof(load_t *) * (cpuids_quantity + 1))) == NULL)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Can't allocate memory for load array!\n");
		return NOTOK;
	}

	/* Get all mastercpuids */
	if ((*ptr_cpu_get_mastercpuids)(cpuids, &mastercpuids, &mastercpuids_quantity) != OK)
		return NOTOK;

	i = 0;
	counter = 0;
	while (mastercpuids[i] >= 0)
	{
		/* Get affected_cpuids */
		if ((*ptr_cpu_get_affectedcpuids)(mastercpuids[i], &affectedcpuids, &affectedcpuids_quantity) != OK)
			return NOTOK;

		j = 0;
		while (affectedcpuids[j] >= 0)
		{
			if ((load = (load_t *)malloc(sizeof(load_t))) == NULL)
			{
				libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Can't allocate memory for load struct!\n");
				return NOTOK;
			}
			memset(load, 0, sizeof(load_t));
			load->mastercpuid = mastercpuids[i];
			load->cpuid = affectedcpuids[j];
			load->total = 0;
			load->usage = 0;

			loads[counter] = load;

			if (counter > cpuids_quantity)
			{
				libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Quantity of all affected cpuids(%d) is to high!\n", counter);
				return NOTOK;
			}

			counter++;
			j++;
		}
		free(affectedcpuids);

		i++;
	}
	loads[counter] = NULL;

	free(mastercpuids);
	free(cpuids);

	if (counter != cpuids_quantity)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Quantity of all affected cpuids(%d) is different to quantity of all cpuids(%d)!\n", counter, cpuids_quantity);
		return NOTOK;
	}

	/* Set load for every cpuid */
	i = 0;
	while (loads[i] != NULL)
	{
		if ((*ptr_cpu_get_load)(loads[i]->cpuid, &loads[i]->total, &loads[i]->usage, &load_tmp) != OK)
			return NOTOK;

		i++;
	}

	return OK;
}


/* Clean up */
int cpu_tini(void)
{
	int i;

	if ((*ptr_cpu_tini)() != OK)
		return NOTOK;

	i = 0;
	while (loads[i] != NULL)
	{
		free(loads[i]);
		i++;
	}

	free(loads);

	return OK;
}


/* Get mastercpuid load */
int cpu_get_load(int mastercpuid, float *load)
{
	int i;
	int cpuids_quantity;
	float load_cpuid;

	i = 0;
	*load = 0.0;
	cpuids_quantity = 0;
	while (loads[i] != NULL)
	{
		if (loads[i]->mastercpuid == mastercpuid)
		{
			if ((*ptr_cpu_get_load)(loads[i]->cpuid, &loads[i]->total, &loads[i]->usage, &load_cpuid) != OK)
				return NOTOK;

			if (load_cpuid > 1.0)
				load_cpuid = 1.0;

			/* Max cpu load per mastercpu will be used */
			if (*load < load_cpuid)
				*load = load_cpuid;

			cpuids_quantity++;
		}

		i++;
	}
	
	if (cpuids_quantity < 1)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "No cpuids for mastercpuid %d found!\n", mastercpuid);
		return NOTOK;
	}

	return OK;
}


/* Get laptop_mode */
int cpu_get_laptop_mode(int *laptop_mode)
{
	return (*ptr_cpu_get_laptop_mode)(laptop_mode);
}


/* Get mastercpuids */
int cpu_get_mastercpuids(int **mastercpuids, int *quantity)
{
	int i;
	int mastercpuid;
	int cpuids_quantity;

	/* Alloc memory for master cpuids */
	cpuids_quantity = 0;
	while (loads[cpuids_quantity] != NULL)
		cpuids_quantity++;

	if ((*mastercpuids = malloc((cpuids_quantity + 1) * sizeof(int))) == NULL)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Can't allocate memory for master cpuids!\n");
		return NOTOK;
	}

	/* Set master cpuids */
	i = 0;
	*quantity = 0;
	mastercpuid = -1;
	while (loads[i] != NULL)
	{
		if (mastercpuid != loads[i]->mastercpuid)
		{
			mastercpuid = loads[i]->mastercpuid;

			(*mastercpuids)[*quantity] = loads[i]->mastercpuid;
			(*quantity)++;
		}

		i++;
	}
	(*mastercpuids)[*quantity] = -1;

	if (*quantity < 1)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "No master cpuids found!\n");
		return NOTOK;
	}

	return OK;
}


/* Get affected cpuids */
int cpu_get_affectedcpuids(int mastercpuid, int **affectedcpuids, int *quantity)
{
	int i;
	int cpuids_quantity;

	/* Alloc memory for affected cpuids */
	cpuids_quantity = 0;
	while (loads[cpuids_quantity] != NULL)
		cpuids_quantity++;

	if ((*affectedcpuids = malloc((cpuids_quantity + 1) * sizeof(int))) == NULL)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Can't allocate memory for affected cpuids!\n");
		return NOTOK;
	}

	/* Set affected cpuids */
	i = 0;
	*quantity = 0;
	while (loads[i] != NULL)
	{
		if (mastercpuid == loads[i]->mastercpuid)
		{
			(*affectedcpuids)[*quantity] = loads[i]->cpuid;
			(*quantity)++;
		}

		i++;
	}
	(*affectedcpuids)[*quantity] = -1;

	if (*quantity < 1)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "No affected cpuids found!\n");
		return NOTOK;
	}

	return OK;
}


/* Get frequencies */
int cpu_get_frequencies(int mastercpuid, int **frequencies, int *quantity)
{
	int i;
	int index;

	if ((*ptr_cpu_get_frequencies)(mastercpuid, frequencies, quantity) != OK)
		return NOTOK;

	/* Sort the frequencies in ascending order */
	qsort(*frequencies, *quantity, sizeof(int), &libcpupowerd_compare_ints);

	/* Delete multiple frequency entries */
	if (*quantity > 1)
	{
		i = 1;
		index = 1;
		while ((*frequencies)[i] >= 0)
		{
			if ((*frequencies)[i - 1] != (*frequencies)[i])
			{
				(*frequencies)[index] = (*frequencies)[i];
				index++;
			}

			i++;
		}

		(*frequencies)[index] = -1;
		*quantity = index;
	}

	return OK;
}


/* Get current frequency */
int cpu_get_frequency(int mastercpuid, int *frequency)
{
	return (*ptr_cpu_get_frequency)(mastercpuid, frequency);
}


/* Set current frequency */
int cpu_set_frequency(int mastercpuid, int frequency)
{
	return (*ptr_cpu_set_frequency)(mastercpuid, frequency);
}


/* Get current governor */
int cpu_get_governor(int mastercpuid, char *governor, int size)
{
	return (*ptr_cpu_get_governor)(mastercpuid, governor, size);
}


/* Set current governor */
int cpu_set_governor(int mastercpuid, char *governor)
{
	return (*ptr_cpu_set_governor)(mastercpuid, governor);
}





