/************************************************************************/
/* Copyright (C) 2008-2009 Markus Strobl <markus.strobl@gmx.at>         */
/* Copyright (c) 2008 Bernhard Froehlich <decke@bluelife.at>            */
/*                                                                      */
/* cpupowerd is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by    */
/* the Free Software Foundation; either version 2 of the License.       */
/*                                                                      */
/* cpupowerd is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of           */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */
/*                                                                      */
/* You should have received a copy of the GNU General Public License    */
/* along with GNU Emacs; see the file COPYING.  If not, write to        */
/* the Free Software Foundation, Inc., 59 Temple Place - Suite 330,     */
/* Boston, MA 02111-1307, USA.                                          */
/************************************************************************/
/* $Id: cpu_bsd.c 369 2009-01-01 13:05:07Z markus $                     */
/************************************************************************/


/* Includes */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/param.h>
#include <sys/resource.h>
#include <sys/sysctl.h>

#include "cpupowerd.h"
#include "libcpupowerd.h"
#include "cpu_bsd.h"


/* Defines */
#define	FREQUENCYROUND(frequency)	((int)((frequency / 25.0) + 0.5)) * 25
#define FREEBSD_GOVERNOR			"userspace"
#define SYSCTL_NUMCPUS				"hw.ncpu"
#define SYSCTL_MAXCPUS				"kern.smp.maxcpus"
#define SYSCTL_CPUTIME				"kern.cp_time"
#define SYSCTL_CPUTIMES				"kern.cp_times"
#define SYSCTL_ACLINE				"hw.acpi.acline"


/* Prototypes */
int cpu_bsd_tini(void);
int cpu_bsd_get_load(int mastercpuid, unsigned long long *total, unsigned long long *usage, float *load);
int cpu_bsd_get_laptop_mode(int *laptop_mode);
int cpu_bsd_get_mastercpuids(int *cpuids, int **mastercpuids, int *quantity);
int cpu_bsd_get_affectedcpuids(int mastercpuid, int **affectedcpuids, int *quantity);
int cpu_bsd_get_frequencies(int mastercpuid, int **frequencies, int *quantity);
int cpu_bsd_get_frequency(int mastercpuid, int *frequency);
int cpu_bsd_set_frequency(int mastercpuid, int frequency);
int cpu_bsd_get_governor(int mastercpuid, char *governor, int size);
int cpu_bsd_set_governor(int mastercpuid, char *governor);
static int cpu_bsd_is_governor(int mastercpuid, char *governor);
static int cpu_bsd_get_frequencies_bsd(int mastercpuid, int **frequencies, int *quantity);


/* Functions */
/* Init */
int cpu_bsd_init(int (**ptr_cpu_get_load) (int, unsigned long long *, unsigned long long *, float *),
				   int (**ptr_cpu_get_laptop_mode) (int *),
				   int (**ptr_cpu_get_mastercpuids) (int *, int **, int *),
				   int (**ptr_cpu_get_affectedcpuids) (int, int **, int *),
				   int (**ptr_cpu_get_frequencies) (int, int **, int *),
				   int (**ptr_cpu_get_frequency) (int, int *),
				   int (**ptr_cpu_set_frequency) (int, int),
				   int (**ptr_cpu_get_governor) (int, char *, int),
				   int (**ptr_cpu_set_governor) (int, char *),
				   int (**ptr_cpu_tini) (void))
{
	*ptr_cpu_get_load = cpu_bsd_get_load;
	*ptr_cpu_get_laptop_mode = cpu_bsd_get_laptop_mode;
	*ptr_cpu_get_mastercpuids = cpu_bsd_get_mastercpuids;
	*ptr_cpu_get_affectedcpuids = cpu_bsd_get_affectedcpuids;
	*ptr_cpu_get_frequencies = cpu_bsd_get_frequencies;
	*ptr_cpu_get_frequency = cpu_bsd_get_frequency;
	*ptr_cpu_set_frequency = cpu_bsd_set_frequency;
	*ptr_cpu_get_governor = cpu_bsd_get_governor;
	*ptr_cpu_set_governor = cpu_bsd_set_governor;
	*ptr_cpu_tini = cpu_bsd_tini;

	return OK;
}


/* Clean up */
int cpu_bsd_tini(void)
{
	return OK;
}


/* Get load for mastercpuid */
/* Format: <user> <nice> <sys> <interrupt> <idle> */
int cpu_bsd_get_load(int cpuid, unsigned long long *total, unsigned long long *usage, float *load)
{
	int index;
	int numcpus;
	int maxcpus;
	int multifactor;
	char sysctl_cptime_string[32];
	long *cpu_times;
	size_t cpu_time_size;
	size_t cpu_time_size_marker;
	unsigned long long total_current;
	unsigned long long usage_current;

	if (libcpupowerd_getsysctl(SYSCTL_MAXCPUS, (void *)&maxcpus, sizeof(maxcpus)) != OK)
		return NOTOK;

	if (libcpupowerd_getsysctl(SYSCTL_NUMCPUS, (void *)&numcpus, sizeof(numcpus)) != OK)
		return NOTOK;

	#if __FreeBSD_version < 700103
	multifactor = 0;
	cpu_time_size = (size_t)(sizeof(long) * CPUSTATES);
	cpu_time_size_marker = cpu_time_size;
	snprintf(sysctl_cptime_string, sizeof(sysctl_cptime_string), "%s", SYSCTL_CPUTIME);
	#else
	multifactor = CPUSTATES;
	cpu_time_size = (size_t)(sizeof(long) * CPUSTATES * maxcpus);
	cpu_time_size_marker = cpu_time_size;
	snprintf(sysctl_cptime_string, sizeof(sysctl_cptime_string), "%s", SYSCTL_CPUTIMES);
	#endif

	if ((cpu_times = (long *)malloc(cpu_time_size)) == NULL)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Can't allocate memory for cpu_times array!\n");
      	return NOTOK;
	}

	if (libcpupowerd_getsysctl(sysctl_cptime_string, (void *)cpu_times, cpu_time_size) != OK)
		return NOTOK;

	index = cpuid * multifactor;

	/* cpu_time_size is lower than index -> error */
	if (index + CPUSTATES >= cpu_time_size)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Index(%d) for cp_times(%d) is to high!\n", index, cpu_time_size);
		return NOTOK;
	}

	total_current = cpu_times[index + CP_USER] + cpu_times[index + CP_NICE] + cpu_times[index + CP_SYS] + cpu_times[index + CP_INTR] + cpu_times[index + CP_IDLE];
	usage_current = cpu_times[index + CP_USER] + cpu_times[index + CP_NICE] + cpu_times[index + CP_SYS] + cpu_times[index + CP_INTR];

	*load = (float)(usage_current - *usage) / (float)(total_current - *total);

	/* Workarround to get 100% load per cpu like linux */
	*load *= numcpus;

	*total = total_current;
	*usage = usage_current;

	free(cpu_times);

	return OK;
}


/* Get laptop mode, "laptop_mode" on FreeBSD is only a check to the AC Line */
int cpu_bsd_get_laptop_mode(int *laptop_mode)
{	
	int acline;
	
	acline = 1;
	if (libcpupowerd_getsysctl(SYSCTL_ACLINE, (void *)&acline, sizeof(acline)) != OK)
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_INFO, "Laptop mode can't be detected, using laptop mode 0!\n");
	
	if (acline == 0)
		*laptop_mode = 1;
	else
		*laptop_mode = 0;
		
	return OK;
}


/* Get cpu mastercpuids */
int cpu_bsd_get_mastercpuids(int *cpuids, int **mastercpuids, int *quantity)
{
	int i;
	int quantity_cpuids;
	size_t length;
	int levels_mib[4];
	char levels_sysctl[32];

	quantity_cpuids = 0;
	while(cpuids[quantity_cpuids] >= 0)
		quantity_cpuids++;

	/* Allocate mastercpuids array with the quantity of cpuids, this is more than enough */
	if ((*mastercpuids = malloc((quantity_cpuids + 1) * sizeof(int))) == NULL)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Can't allocate memory for mastercpuids array!\n");
        return NOTOK;
	}

	i = 0;
	*quantity = 0;
	while (cpuids[i] >= 0)
	{
		/* if freq_levels is available, cpuid is a mastercpuid */
		length = 4;
		snprintf(levels_sysctl, sizeof(levels_sysctl), "dev.cpu.%d.freq_levels", cpuids[i]);
		if (sysctlnametomib(levels_sysctl, levels_mib, &length) == 0)
		{
			(*mastercpuids)[*quantity] = cpuids[i];
			(*quantity)++;
		}

		i++;
	}
	(*mastercpuids)[*quantity] = -1;

	if (*quantity < 1)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "No master cpuids found!\n");
		return NOTOK;
	}

	return OK;
}


/* Get affected cpuids */
int cpu_bsd_get_affectedcpuids(int mastercpuid, int **affectedcpuids, int *quantity)
{
	int i;
	int maxcpus;
	int affectedcpuids_quantity;
	size_t length;
	int levels_mib[4];
	char levels_sysctl[32];

	/* Check, if mastercpuid is really a mastercpuid */
	length = 4;
	snprintf(levels_sysctl, sizeof(levels_sysctl), "dev.cpu.%d.freq_levels", mastercpuid);
	if (sysctlnametomib(levels_sysctl, levels_mib, &length) != 0)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Cpuid %d isn't a mastercpuid!\n", mastercpuid);
		return NOTOK;
	}

	/* Get max quantity of cpus */
	if (libcpupowerd_getsysctl(SYSCTL_MAXCPUS, (void *)&maxcpus, sizeof(maxcpus)) != OK)
		return NOTOK;

	affectedcpuids_quantity = 1;
	for (i = mastercpuid + 1; i < maxcpus; i++)
	{
		length = 4;
		snprintf(levels_sysctl, sizeof(levels_sysctl), "dev.cpu.%d.freq_levels", i);
		if (sysctlnametomib(levels_sysctl, levels_mib, &length) != 0)
			affectedcpuids_quantity++;
		else
			break;
	}

	/* Allocate affected cpuids array with the count of affectedcpuids_quantitymax, this is more than enough */
	if ((*affectedcpuids = malloc((affectedcpuids_quantity + 1) * sizeof(int))) == NULL)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Can't allocate memory for affected cpuids array!\n");
        return NOTOK;
	}

	/* Set the affected cpuids */
	*quantity = 0;
	(*affectedcpuids)[*quantity] = mastercpuid;
	(*quantity)++;
	for (i = mastercpuid + 1; i < mastercpuid + affectedcpuids_quantity; i++)
	{
		length = 4;
		snprintf(levels_sysctl, sizeof(levels_sysctl), "dev.cpu.%d", i);
		if (sysctlnametomib(levels_sysctl, levels_mib, &length) == 0)
		{
			(*affectedcpuids)[*quantity] = i;
			(*quantity)++;
		}
	}
	(*affectedcpuids)[*quantity] = -1;

	return OK;
}


/* Get cpu frequencies */
int cpu_bsd_get_frequencies(int mastercpuid, int **frequencies, int *quantity)
{
	int i;

	if (cpu_bsd_get_frequencies_bsd(mastercpuid, frequencies, quantity) != OK)
		return NOTOK;

	i = 0;
	while ((*frequencies)[i] >= 0)
	{
		(*frequencies)[i] = FREQUENCYROUND((*frequencies)[i]);
		i++;
	}

	return OK;
}


/* Get frequency */
int cpu_bsd_get_frequency(int mastercpuid, int *frequency)
{
	char frequency_sysctl[32];
	
	snprintf(frequency_sysctl, sizeof(frequency_sysctl), "dev.cpu.%d.freq", mastercpuid);
	
	if (libcpupowerd_getsysctl(frequency_sysctl, (void *)frequency, sizeof(*frequency)) != OK)
		return NOTOK;

	*frequency = FREQUENCYROUND(*frequency);

	return OK;
}


int cpu_bsd_set_frequency(int mastercpuid, int frequency)
{
	int index;
	int found;
	int quantity;
	size_t length = 4;
	int mib[4];
	int *frequencies;
	char frequency_sysctl[32];

	if (cpu_bsd_get_frequencies_bsd(mastercpuid, &frequencies, &quantity) != OK)
		return NOTOK;

	index = 0;
	found = 0;
	while (frequencies[index] >= 0)
	{
		if (FREQUENCYROUND(frequencies[index]) == frequency)
		{
			found = 1;
			break;
		}

		index++;
	}

	if (found != 1)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Frequency %d MHz for mastercpuid %d not found!\n", frequency, mastercpuid);
		return NOTOK;
	}

	snprintf(frequency_sysctl, sizeof(frequency_sysctl), "dev.cpu.%d.freq", mastercpuid);
	if (sysctlnametomib(frequency_sysctl, mib, &length) != 0)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Can't get mib for %s!\n", frequency_sysctl);
		return NOTOK;
	}

	if (sysctl(mib, 4, NULL, NULL, &frequencies[index], sizeof(frequencies[index])) != 0)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Can't set frequency %d MHz with sysctl!\n", frequencies[index]);
		return NOTOK;
	}
	
	free(frequencies);
        
	return OK;
}


/* Get governor */
int cpu_bsd_get_governor(int mastercpuid, char *governor, int size)
{
	strcpy(governor, FREEBSD_GOVERNOR);

	return OK;
}


/* Set governor */
int cpu_bsd_set_governor(int mastercpuid, char *governor)
{
	char governor_tmp[32];

	/* Compare with current governor */
	if (cpu_bsd_get_governor(mastercpuid, governor_tmp, (int)sizeof(governor_tmp)) != OK)
		return NOTOK;

	if (strcmp(governor, governor_tmp) == 0)
		return OK;

	/* Is governor available */
	if (cpu_bsd_is_governor(mastercpuid, governor) != OK)
		return NOTOK;

	return OK;
}


/* Check if governor is available */
static int cpu_bsd_is_governor(int mastercpuid, char *governor)
{
	if (strcmp(governor, FREEBSD_GOVERNOR) == 0)
		return OK;

	return NOTOK;
}


/* Get the frequencies for a mastercpu */
static int cpu_bsd_get_frequencies_bsd(int mastercpuid, int **frequencies, int *quantity)
{
	int i;
	size_t length;
	int levels_mib[4];
	char levels_sysctl[32];
	char *frequencies_string;
	char *p, *q;

	length = 4;
	snprintf(levels_sysctl, sizeof(levels_sysctl), "dev.cpu.%d.freq_levels", mastercpuid);
	if (sysctlnametomib(levels_sysctl, levels_mib, &length) != 0)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "No frequencies for mastercpuid %d available!\n", mastercpuid);
		return NOTOK;
	}

	length = 0;
	if (sysctl(levels_mib, 4, NULL, &length, NULL, 0) != 0)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Error on calling sysctl!\n");
		return NOTOK;
	}

	if ((frequencies_string = malloc(length)) == NULL)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Can't allocate memory for frequencies string!\n");
		return NOTOK;
	}

	if (sysctl(levels_mib, 4, frequencies_string, &length, NULL, 0) != 0)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Error on calling sysctl!\n");
		return NOTOK;
	}

	if (libcpupowerd_get_integers_from_string(frequencies_string, "%d/%*d", ' ', frequencies, quantity) != LIBCPUPOWERD_OK)
		return NOTOK;

	free(frequencies_string);

	return OK;
}





