/************************************************************************/
/* Copyright (C) 2008-2009 Markus Strobl <markus.strobl@gmx.at>         */
/*                                                                      */
/* cpupowerd is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by    */
/* the Free Software Foundation; either version 2 of the License.       */
/*                                                                      */
/* cpupowerd is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of           */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */
/*                                                                      */
/* You should have received a copy of the GNU General Public License    */
/* along with GNU Emacs; see the file COPYING.  If not, write to        */
/* the Free Software Foundation, Inc., 59 Temple Place - Suite 330,     */
/* Boston, MA 02111-1307, USA.                                          */
/************************************************************************/
/* $Id: cpuid.c 369 2009-01-01 13:05:07Z markus $                       */
/************************************************************************/


/* Includes */
#include <string.h>

#include "cpuid.h"
#include "cpupowerd.h"
#include "libcpupowerd.h"
#ifdef __linux__
#include "cpuid_linux.h"
#elif __FreeBSD__
#include "cpuid_bsd.h"
#endif


/* Function pointers */
int (*ptr_cpuid_get_cpuids) (int **, int *);
int (*ptr_cpuid_get_oldestmodel) (int, int, int *);
int (*ptr_cpuid_tini) (void);


/* Prototypes */
static int cpuid_modify_model(int vendor, int family, int *model);
static int cpuid_get_vendor(int *vendor);
static int cpuid_cpuid(unsigned int op, unsigned long *eax, unsigned long *ebx, unsigned long *ecx, unsigned long *edx);


/* Functions */
/* Init */
int cpuid_init(void)
{
	int result;

	#ifdef __linux__
	result = cpuid_linux_init(&ptr_cpuid_get_cpuids, &ptr_cpuid_get_oldestmodel, &ptr_cpuid_tini);
	#elif __FreeBSD__
	result = cpuid_bsd_init(&ptr_cpuid_get_cpuids, &ptr_cpuid_get_oldestmodel, &ptr_cpuid_tini);
	#else
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Your operating system isn't yet supported!\n");
		result = NOTOK;
	#endif

	return result;
}


/* Clean up */
int cpuid_tini(void)
{
	return (*ptr_cpuid_tini)();
}


/* Get the cpuids */
int cpuid_get_cpuids(int **cpuids, int *quantity)
{
	return (*ptr_cpuid_get_cpuids)(cpuids, quantity);
}


/* Get the cpu info */
int cpuid_get_cpuinfo(int *vendor, int *family, int *model)
{
	int model_tmp;
	int extended_model_tmp;
	int extended_family_tmp;
	int family_tmp;
	unsigned long eax, ebx, ecx, edx;

	if (cpuid_get_vendor(vendor) != OK)
		return NOTOK;

	if (cpuid_cpuid(0x1, &eax, &ebx, &ecx, &edx) != OK)
		return NOTOK;

	/* For model, extended model, family and extended family 
       have a look at www.sandpile.org/ia32/cpuid.htm */
	model_tmp = (eax >> 4) & 0xF;
	extended_model_tmp = (eax >> 16) & 0xF;
	family_tmp = (eax >> 8) & 0xF;
	extended_family_tmp = (eax >> 20) & 0xFF;

	switch (*vendor)
	{
		case VENDOR_AMD:

			switch (family_tmp)
			{
				case 0x0:	/* Refer to extended family */
				case 0xF:	/* Refer to extended family */

					switch (extended_family_tmp)
					{
						case 0x00:	/* AMD K8 */
							*family = AMD_K8;
							*model = extended_model_tmp;
							break;

						case 0x01:	/* AMD K8L also known as AMD K10 */
							*family = AMD_K10;
							*model = extended_model_tmp;
							break;

						default:
							libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Your AMD CPU isn't supported yet (cpuid: 0x%X)!\n", eax);
							return NOTOK;
					}

					break;

				default:
					libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Your AMD CPU isn't supported (cpuid: 0x%X)!\n", eax);
					return NOTOK;
			}

			break;

		case VENDOR_INTEL:

			switch (family_tmp)
			{
				case 0x6:	/* Intel P6, P2, P3, PM, Core 2 */

					switch (model_tmp)
					{
						case 0x9:	/* PM (0.13 µm) with 1 MB o-die L2 */
						case 0xD:	/* PM (0.09 µm) with 2 MB on-die L2 */
							*family = INTEL_PENTIUM_M;
							*model = model_tmp;
							break;

						case 0xE:	/* PM DC (65 nm) with 2 MB on-die L2 */
							*family = INTEL_CORE;
							*model = model_tmp;
							break;

						case 0xF:	/* Core 2 DC (65 nm) with 4 MB on-die L2 */
						case 0x16:	/* Core 2 SC (65 nm) with 1 MB on-die L2 */
						case 0x17:	/* Core 2 DC (45 nm) with 6 MB on-die L2 */
							*family = INTEL_CORE_2;
							*model = model_tmp;
							break;

						default:
							libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Your Intel CPU isn't supported yet (cpuid: 0x%X)!\n", eax);
							return NOTOK;
					}

					break;

				case 0x0:	/* Refer to extended family */
				case 0xF:	/* Refer to extended family */

					switch (extended_model_tmp)
					{
						case 0x1:	/* Intel Core 2: 65 nm SC with 1 MB on-die L2, 45 nm DC with 6 MB on-die L2 */
							*family = INTEL_CORE_2;
							*model = extended_model_tmp;
							break;

						default:
							libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Your Intel CPU isn't supported yet (cpuid: 0x%X)!\n", eax);
							return NOTOK;
					}

					break;

				default:
					libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Your Intel CPU isn't supported yet! (cpuid: 0x%X)\n", eax);
					return NOTOK;
			}

			break;

		default:
			libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Your CPU vendor isn't supported!\n");
			return NOTOK;
	}

	if (cpuid_modify_model(*vendor, *family, model) != OK)
		return NOTOK;

	return OK;
}


/* Modify model, if dangerous cpu mix (use the oldest model) */
static int cpuid_modify_model(int vendor, int family, int *model)
{
	if (vendor != VENDOR_AMD && family != AMD_K8)
		return OK;

	return (*ptr_cpuid_get_oldestmodel)(vendor, family, model);
}


/* Get vendor */
static int cpuid_get_vendor(int *vendor)
{
	unsigned long eax, ebx, ecx, edx;
	char cpuvendor[16];

	if (cpuid_cpuid(0x0, &eax, &ebx, &ecx, &edx) != OK)
		return NOTOK;

	memset(cpuvendor, '\0', sizeof(cpuvendor));
	memcpy(&cpuvendor[0], &ebx, 4);
	memcpy(&cpuvendor[4], &edx, 4);
	memcpy(&cpuvendor[8], &ecx, 4);
	cpuvendor[13] = '\0';

	if (strncmp(cpuvendor,"AuthenticAMD", sizeof(cpuvendor)) == 0)
		*vendor = VENDOR_AMD;
	else if (strncmp(cpuvendor, "GenuineIntel", sizeof(cpuvendor)) == 0)
		*vendor = VENDOR_INTEL;
	else 
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Your cpu vendor isn't supported yet!\n");
    	return NOTOK;
	}

	return OK;
}


/* Reads cpuid information with assembler */
static int cpuid_cpuid(unsigned int op, unsigned long *eax, unsigned long *ebx, unsigned long *ecx, unsigned long *edx)
{
	asm("cpuid"
		: "=a" (*eax),
		  "=b" (*ebx),
		  "=c" (*ecx),
		  "=d" (*edx)
		: "0" (op));

	return OK;
}






