/************************************************************************/
/* Copyright (C) 2008-2009 Markus Strobl <markus.strobl@gmx.at>         */
/*                                                                      */
/* cpupowerd is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by    */
/* the Free Software Foundation; either version 2 of the License.       */
/*                                                                      */
/* cpupowerd is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of           */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */
/*                                                                      */
/* You should have received a copy of the GNU General Public License    */
/* along with GNU Emacs; see the file COPYING.  If not, write to        */
/* the Free Software Foundation, Inc., 59 Temple Place - Suite 330,     */
/* Boston, MA 02111-1307, USA.                                          */
/************************************************************************/
/* $Id: cpupowerd.c 369 2009-01-01 13:05:07Z markus $                   */
/************************************************************************/


/* Includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <signal.h>

#include "cpu.h"
#include "msr.h"
#include "cpuid.h"
#include "cpupowerd.h"
#include "libcpupowerd.h"


/* Defines */
#define NAME								"cpupowerd"
#define MAJOR 								0
#define MINOR 								2
#define UPDATE 								1
#define AUTHOR								"Markus Strobl"
#define EMAIL								"markus.strobl@gmx.at"

#define ENVIRONMENT_INVALID_COUNTER_MAX		10
#define ENVIRONMENT_INVALID_TIME_PERIOD		3600

#define SPEED_UP_INIT 						80
#define SPEED_DOWN_INIT 					20
#define LOOP_TIME 							1000
#define GOVERNOR							"userspace"
#define REDUCE_VOLTAGE_MAX					0.2
#define ENHANCE_VOLTAGE_MAX					0.1


/* Modes of changing cpu frequency */
enum modes 
{
	LOWER,
	EQUAL,
	UPPER
};


/* Strategies of changing cpu frequency */
enum strategies
{
	MINIMUM,
	SINE,
	AGGRESSIVE,
	PASSIVE,
	BSD,
	LEAPS,
	MAXIMUM
};


/* Step for mastercpu */
typedef struct step
{
	int fid;										/* FID of the step */
	int vid;										/* VID of the step */
	int vid_orig;									/* Original vid of this step */
	int is_set_from_configfile;						/* Flag, if step is set from config file */
} step_t;


/* Mastercpus for all physical cpus */
typedef struct mastercpu
{
	int cpuid;										/* Cpuid */
	int step_index;									/* Current index of step array */
	int step_index_max;								/* Maximal index of step array */
	char governor_default[32];						/* Original governor */
	step_t **steps;									/* Array of possible steps for this mastercpu in ascending order */
} mastercpu_t;


/* Global variables */
mastercpu_t **mastercpus;							/* Global array that contains a pointer to all mastercpuids. */
int environment_invalid_counter = 0;				/* If governor were adjusted from other program, situation will be cleared and environment_invalid_counter increased. */
unsigned long environment_invalid_timestamp = 0;	/* Timestamp of the environment invalid condition */
char *p_configfile = NULL;							/* Filename for config file */
char *p_autoconfigfile = NULL;						/* Filename for autoconfig file */
int p_program_work_mode = 0;						/* Workmode of the program, default 0 foreground, 1 daemon */
int p_laptop = 0;									/* Consider laptop accu operating */
int p_status = 0;									/* Show status of mastercpus */
int p_time = LOOP_TIME;								/* Interval to look usage of cpu in fore or background mode */
int p_voltage_cpuid = -1;							/* Mastercpuid which should change the voltage */
float p_reduce_voltage = 0.0;						/* Reduce voltage to original voltage for all mastercpus */
float p_enhance_voltage = 0.0;						/* Enhance voltage to original voltage for all mastercpus */
float p_voltage = -1.0;								/* Voltage value */
int p_frequency_cpuid = -1;							/* Mastercpuid which should change the frequency */
int p_frequency = -1;								/* Frequency value */
int p_speed_up = SPEED_UP_INIT;						/* Threshold up */
int p_speed_down = SPEED_DOWN_INIT;					/* Threshold down */
int p_message_level = LIBCPUPOWERD_MESSAGE_WARNING;	/* Output message level */
enum strategies p_strategy = BSD;					/* Strategy for up and down frequencies */


/* Prototypes */
int is_mastercpuid(int mastercpuid);
int is_fid(int cpuid, int fid, int *mastercpu_index, int *step_index);
int get_fid(int cpuid, int *fid);
int set_fid(int cpuid, int fid);
int set_vid(int cpuid, int vid);
int read_config(char *file);
int write_config(char *file);
int set_config(int cpuid, int fid, int vid);
int reorganize_frequencies(void);
int is_undervolting_environment_valid(mastercpu_t *mastercpu);
int get_vid_with_offset(int vid, int *vid_with_offset);
int change_cpustep(mastercpu_t * mastercpu, enum modes mode);
int decide_change(mastercpu_t *mastercpu, enum modes *mode);
int tini(void);
int init(void);
int show_status(void);
void show_exitmessage(char *message);
int show_version(void);
int show_help(void);
int parse_options(int argc, char **argv);
void terminate(int signum);


/* Functions */
/* Check if cpuid is valid */
int is_mastercpuid(int cpuid)
{
	int i;
	int found;

	i = 0;
	found = 0;
	while (mastercpus[i] != NULL)
	{
		if (mastercpus[i]->cpuid == cpuid)
		{
			found = 1;
			break;
		}

		i++;
	}
	
	if (found == 0)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Cpuid %d isn't a mastercpuid!\n", cpuid);
		return NOTOK;
	}
	
	return OK;
}


/* Check if fid is valid */
int is_fid(int cpuid, int fid, int *mastercpu_index, int *step_index)
{
	int i, j;
	int found;
	int frequency;

	i = 0;
	found = 0;
	while (mastercpus[i] != NULL)
	{
		j = 0;
		while (mastercpus[i]->steps[j] != NULL)
		{
			if (mastercpus[i]->steps[j]->fid == fid)
			{
				found = 1;
				*mastercpu_index = i;
				*step_index = j;
				break;
			}
			
			j++;
		}

		if (found != 0)
			break;

		i++;
	}
	
	if (found == 0)
	{
		if (msr_fid_to_frequency(fid, &frequency) != OK)
			return NOTOK;

		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Frequency %d MHz not available!\n", frequency);
		return NOTOK;
	}
	
	return OK;
}


/* Get fid */
int get_fid(int cpuid, int *fid)
{
	int frequency;

	if (cpu_get_frequency(cpuid, &frequency) != OK)
		return NOTOK;

	if (msr_frequency_to_fid(frequency, fid) != OK)
		return NOTOK;

	return OK;
}


/* Set fid */
int set_fid(int cpuid, int fid)
{
	int fid_tmp;
	int vid_tmp;
	int frequency;
	int frequency_tmp;
	int step_index;
	int mastercpu_index;

	if (is_mastercpuid(cpuid) != OK)
		return NOTOK;

	if (is_fid(cpuid, fid, &mastercpu_index, &step_index) != OK)
		return NOTOK;

	if (msr_fid_to_frequency(fid, &frequency) != OK)
		return NOTOK;

	if (cpu_set_frequency(cpuid, frequency) != OK)
		return NOTOK;

	/* Set current steps index */
	mastercpus[mastercpu_index]->step_index = step_index;

	/* Get vid and fid after frequency changeing */
	if (msr_get_status(cpuid, &vid_tmp, &fid_tmp) != OK)
		return NOTOK;

	if (fid != fid_tmp)
	{
		if (msr_fid_to_frequency(fid_tmp, &frequency_tmp) != OK)
			return NOTOK;

		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "CPU frequency should be %d MHz, but is %d MHz. Frequency changing failed!\n", frequency, frequency_tmp);
		return NOTOK;
	}

	/* Set original cpu voltage for current frequency */
	mastercpus[mastercpu_index]->steps[mastercpus[mastercpu_index]->step_index]->vid_orig = vid_tmp;

	return OK;
}


/* set vid, is only possible by constant frequency */
int set_vid(int cpuid, int vid)
{
	int fid;
	int vid_tmp;

	if (is_mastercpuid(cpuid) != OK)
		return NOTOK;

	if (msr_get_status(cpuid, &vid_tmp, &fid) != OK)
		return NOTOK;

	if (msr_set_status(cpuid, vid, fid) != OK)
		return NOTOK;

	return OK;
}


/* Read the config file */
int read_config(char *file)
{
	int vid;
	int fid;
	int cpuid;
	int counter;
 	int frequency;
 	float voltage;
 	struct stat file_stat;
	FILE *fd;

	if (stat(file, &file_stat) != 0)
 	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Config file %s not available!\n", file);
 		return NOTOK;
 	}

	if ((fd = fopen(file, "r")) == NULL)
 	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Can't open %s for reading!\n", file);
 		return NOTOK;
 	}

	cpuid = -1;
	counter = 0;
	while (feof(fd) == 0)
	{
		if (fscanf(fd, "[%d]\n", &cpuid) != 0)
			continue;

		if (fscanf(fd, "%d %f\n", &frequency, &voltage) != 0)
 		{
			libcpupowerd_message(LIBCPUPOWERD_MESSAGE_INFO, "Config file values: frequency %d MHz, voltage: %.4f V\n", frequency, voltage);

			if (msr_frequency_to_fid(frequency, &fid) != OK)
				return NOTOK;

			if (msr_voltage_to_vid(voltage, &vid) != OK)
				return NOTOK;

			if (set_config(cpuid, fid, vid) != OK)
				return NOTOK;
			else
				counter++;
		}
		else
		{
			while (fgetc(fd) != '\n' && feof(fd) == 0);
				continue;
		}
	}

	if (fclose(fd) != 0)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Can't close file %s!\n", file);
		return NOTOK;
	}

	if (counter == 0)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "No valid lines found in configfile %s!\n", file);
		return NOTOK;
	}

	if (reorganize_frequencies() != OK)
		return NOTOK;

	return OK;
}


/* Write the config file */
int write_config(char *file)
{
	int i, j;
	int step_index;
	int frequency;
	int mastercpus_counter;
	int mastercpus_different;
	float voltage;
	char buffer[32];
	FILE *fd;

	/* Get original vids for frequencies */
	i = 0;
	while (mastercpus[i] != NULL)
	{
		step_index = mastercpus[i]->step_index;

		for (j = 0; j <= mastercpus[i]->step_index_max; j++)
		{
			if (set_fid(mastercpus[i]->cpuid, mastercpus[i]->steps[j]->fid) != OK)
				return NOTOK;
		}
			
		/* Restore frequency */
		if (set_fid(mastercpus[i]->cpuid, mastercpus[i]->steps[step_index]->fid) != OK)
			return NOTOK;

		i++;
	}

	/* Are mastercpu frequencies different */
	mastercpus_counter = i;
	mastercpus_different = 0;
	for (i = 0; i < (mastercpus_counter - 1); i++)
	{
		if (mastercpus[i]->step_index_max != mastercpus[i + 1]->step_index_max)
		{
			mastercpus_different = 1;
			break;
		}

		for (j = 0; j <= mastercpus[i]->step_index_max; j++)
		{
			if (mastercpus[i]->steps[j]->fid != mastercpus[i + 1]->steps[j]->fid ||
                mastercpus[i]->steps[j]->vid_orig != mastercpus[i + 1]->steps[j]->vid_orig)
			{
				mastercpus_different = 1;
				break;
			}
		}

		if (mastercpus_different != 0)
			break;
	}

	/* Write config file */
	if ((fd = fopen(file, "w")) == NULL)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Can't open %s for writing!\n", file);
		return NOTOK;
	}

	i = 0;
	while (mastercpus[i] != NULL)
	{
		/* If frequencies different, write for every mastercpu separat values */
		if (mastercpus_different != 0)
		{
			snprintf(buffer, sizeof(buffer), "[%d]\n", mastercpus[i]->cpuid);
			if (fwrite(buffer, sizeof(char), strlen(buffer), fd) != strlen(buffer))
			{
				libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Can't write mastercpuid %d in %s!\n", mastercpus[i]->cpuid, file);
				return NOTOK;
			}
		}

		j = 0;
		while (mastercpus[i]->steps[j] != NULL)
		{
			if (msr_vid_to_voltage(mastercpus[i]->steps[j]->vid_orig, &voltage) != OK)
				return NOTOK;

			if (msr_fid_to_frequency(mastercpus[i]->steps[j]->fid, &frequency) != OK)
				return NOTOK;
 
			snprintf(buffer, sizeof(buffer), "%d %.4f\n", frequency, voltage);
			if (fwrite(buffer, sizeof(char), strlen(buffer), fd) != strlen(buffer))
			{
				libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Can't write frequency and voltage in %s!\n", file);
				return NOTOK;
			}

			j++;
		}

		if (mastercpus_different == 0)
			break;

		i++;
	}
	
	if (fclose(fd) != 0)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Can't close file %s!\n", file);
		return NOTOK;
	}

	return OK;
}


/* set vid from the config file to the global mastercpu struct */
int set_config(int cpuid, int fid, int vid)
{
	int i, j;
	int found;
	int frequency;

	i = 0;
	found = 0;
	while (mastercpus[i] != NULL)
	{
		if ((cpuid >= 0 && mastercpus[i]->cpuid == cpuid) || cpuid < 0)
		{
			j = 0;
			while (mastercpus[i]->steps[j] != NULL)
			{
				if (mastercpus[i]->steps[j]->fid == fid)
				{
					mastercpus[i]->steps[j]->vid = vid;
					mastercpus[i]->steps[j]->is_set_from_configfile = 1;

					found = 1;
					break;
				}

				j++;
			}

			if (found == 0)
			{
				if (msr_fid_to_frequency(fid, &frequency) != OK)
					return NOTOK;

				libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Frequency %d MHz not available in global mastercpu struct!\n", frequency);
				return NOTOK;
			}
			else
			{
				if (cpuid >= 0)
					break;
			}
		}

		i++;
	}

	return OK;
}


/* If not all frequencies are used, we must reorganize frequencies for all mastercpus */
int reorganize_frequencies(void)
{
	int i, j ,k;

	i = 0;
	while (mastercpus[i] != NULL)
	{
		k = 0; j = 0;
		while (j <= mastercpus[i]->step_index_max)
		{
			if (mastercpus[i]->steps[j]->is_set_from_configfile != 0)
			{
				if (j != k)
					mastercpus[i]->steps[k] = mastercpus[i]->steps[j];

				k++;
			}

			j++;
		}

		if (k == 0)
		{
			libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "No valid frequency found in frequencies reorganization!\n");
			return NOTOK;
		}
		else if (k != mastercpus[i]->step_index_max + 1)
		{
			libcpupowerd_message(LIBCPUPOWERD_MESSAGE_INFO, "From %d available frequencies are only %d will be used.\n", mastercpus[i]->step_index_max + 1, k);

			/* Set the limits */
			mastercpus[i]->step_index_max = k - 1;
			mastercpus[i]->steps[k] = NULL;

			if (mastercpus[i]->step_index > mastercpus[i]->step_index_max)
				mastercpus[i]->step_index = mastercpus[i]->step_index_max;

			/* Set fid, because the current fid can be eliminated with config file */
			if (set_fid(mastercpus[i]->cpuid, mastercpus[i]->steps[mastercpus[i]->step_index]->fid) != OK)
				return NOTOK;
		}
		else
			libcpupowerd_message(LIBCPUPOWERD_MESSAGE_INFO, "All %d available frequencies are used.\n", mastercpus[i]->step_index_max + 1);
	
		i++;
	}

	return OK;
}


/* Is environment valid for undervolting */
int is_undervolting_environment_valid(mastercpu_t *mastercpu)
{
	int frequency;
	int frequency_current;
	int fid_current;
	int is_environment_invalid;
	char governor[32];
	unsigned long timestamp;

	is_environment_invalid = 0;	

	/* Check governor, because some power manager (e.g. Gnome-Power-Manager) can adjust the governor! */
	if (cpu_get_governor(mastercpu->cpuid, governor, (int)sizeof(governor)) != OK)
		return NOTOK;

	if (strncmp(governor, GOVERNOR, strlen(GOVERNOR)) != 0)
	{
		/* Illegal condition, other program has changed the governor! */
		is_environment_invalid = 1;

		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_WARNING, "Governor %s isn't correct and will be adjusted to %s again!\n", governor, GOVERNOR);

		if (cpu_set_governor(mastercpu->cpuid, GOVERNOR) != OK)
			return NOTOK;
	}

	/* Check fid of the mastercpus again, because other userspace or power manager program can adjust the fids! */
	if (get_fid(mastercpu->cpuid, &fid_current) != OK)
		return NOTOK;

	if (fid_current != mastercpu->steps[mastercpu->step_index]->fid)
	{
		/* Illegal condition, other program has changed the frequency! */
		is_environment_invalid = 1;

		/* Build message */
		if (msr_fid_to_frequency(fid_current, &frequency_current) != OK)
			return NOTOK;

		if (msr_fid_to_frequency(mastercpu->steps[mastercpu->step_index]->fid, &frequency) != OK)
			return NOTOK;

		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_WARNING,
							"Frequency for mastercpu %d isn't correct and will be adjusted from %d to %d again!\n",
							mastercpu->cpuid, frequency_current, frequency);

		/* Set frequency to the last cpupowerd value */
		if (set_fid(mastercpu->cpuid, mastercpu->steps[mastercpu->step_index]->fid) != OK)
			return NOTOK;

		/* Activate undervolting again */
		if (set_vid(mastercpu->cpuid, mastercpu->steps[mastercpu->step_index]->vid) != OK)
			return NOTOK;
	}

	/* If environment invalid is appeared x times in a time window, cpupowerd will be exit */
	if (is_environment_invalid != 0)
	{
		if (libcpupowerd_get_timestamp(&timestamp) != LIBCPUPOWERD_OK)
				return NOTOK;

		/* Reset environment invalid counter if it's time */
		if (timestamp - environment_invalid_timestamp > ENVIRONMENT_INVALID_TIME_PERIOD)
		{
			environment_invalid_counter = 0;
			environment_invalid_timestamp = 0;
		}

		environment_invalid_counter++;

		if (environment_invalid_timestamp == 0)
			environment_invalid_timestamp = timestamp;

		if (environment_invalid_counter >= ENVIRONMENT_INVALID_COUNTER_MAX)
		{
			/* Cpupowerd will be exit now -> the more intelligent gives way */
			libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR,
								"Frequency/governor adjusted %d times in the last %ld seconds -> Exit!\n", 
								environment_invalid_counter, timestamp - environment_invalid_timestamp);

			terminate(0);
		}
	}

	return OK;
}


/* Get vid with offset */
int get_vid_with_offset(int vid, int *vid_with_offset)
{
	int vid_tmp;
	float voltage;

	if (msr_vid_to_voltage(vid, &voltage) != OK)
		return NOTOK;
	
	if (p_reduce_voltage != 0.0)
		voltage -= p_reduce_voltage;
	else if (p_enhance_voltage != 0.0)
		voltage += p_enhance_voltage;
	else
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "No offset voltage available!\n");
		return NOTOK;
	}

	if (msr_voltage_to_vid(voltage, &vid_tmp) != OK)
		return NOTOK;

	*vid_with_offset = vid_tmp;

	return OK;
}


/* Change the frequency and voltage */
int change_cpustep(mastercpu_t *mastercpu, enum modes mode)
{
	int vid_with_offset;
	int frequency;
	int frequency_new;
	int step_index;
	int laptop_mode;
	float voltage;
	float voltage_orig;
  	long seconds;
	long useconds;

	/* Look at mode and strategy */
	step_index = mastercpu->step_index;
	switch (mode)
	{
		case UPPER:
			switch (p_strategy)
			{
				case MINIMUM:
					step_index = 0;
					break;
				case SINE:
				case PASSIVE:
					step_index++;
					break;
				case LEAPS:
				case MAXIMUM:
				case AGGRESSIVE:
					step_index = mastercpu->step_index_max;
					break;
				case BSD:
					step_index += 2;
					break;
				default:
					libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Unknown strategy (%d)!\n", p_strategy);
					return NOTOK;
			}
			break;
		case LOWER:
			switch (p_strategy)
			{
				case MINIMUM:
				case PASSIVE:
				case LEAPS:
					step_index = 0;
					break;
				case BSD:
				case SINE:
				case AGGRESSIVE:
					step_index--;
					break;
				case MAXIMUM:
					step_index = mastercpu->step_index_max;
					break;
				default:
					libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Unknown strategy (%d)!\n", p_strategy);
					return NOTOK;
			}
			break;
		default:
			libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Wrong mode (%d) for frequency change strategy!\n", mode);
			return NOTOK;
	}

	if (p_laptop != 0)
	{
		if (cpu_get_laptop_mode(&laptop_mode) != OK)
			return NOTOK;

		if (laptop_mode > 0)
			step_index = 0;
	}

	/* Is step_index in the allowed band */
	if (step_index < 0)
		step_index = 0;
	if (step_index > mastercpu->step_index_max)
		step_index = mastercpu->step_index_max;

	/* Is environment for undervolting ok */
	if (is_undervolting_environment_valid(mastercpu) != OK)
		return NOTOK;

	/* Is frequency change necessary */
	if (step_index == mastercpu->step_index)
		return OK;

	libcpupowerd_set_timeofday();

	/* Build message for voltage info */
	if (msr_vid_to_voltage(mastercpu->steps[mastercpu->step_index]->vid, &voltage) != OK)
		return NOTOK;

	if (msr_vid_to_voltage(mastercpu->steps[mastercpu->step_index]->vid_orig, &voltage_orig) != OK)
		return NOTOK;

	libcpupowerd_message(LIBCPUPOWERD_MESSAGE_INFO, "Restoring default cpu voltage from %.4f V to %.4f V.\n", voltage, voltage_orig);

	/* Restore original VID for the current frequency */
	if (set_vid(mastercpu->cpuid, mastercpu->steps[mastercpu->step_index]->vid_orig) != OK)
		return NOTOK;

	libcpupowerd_message(LIBCPUPOWERD_MESSAGE_INFO, "Restoring default cpu voltage done.\n");

	/* Build message for frequency info */
	if (msr_fid_to_frequency(mastercpu->steps[mastercpu->step_index]->fid, &frequency) != OK)
		return NOTOK;

	if (msr_fid_to_frequency(mastercpu->steps[step_index]->fid, &frequency_new) != OK)
		return NOTOK;

	libcpupowerd_message(LIBCPUPOWERD_MESSAGE_INFO, "Setting frequency from %d MHz to %d MHz.\n", frequency, frequency_new);

	/* Later, frequency setting would be done over the msr without restoring original vid! */
	if (set_fid(mastercpu->cpuid, mastercpu->steps[step_index]->fid) != OK)
		return NOTOK;

	libcpupowerd_message(LIBCPUPOWERD_MESSAGE_INFO, "Setting frequency done.\n");

	/* Use offset voltage if available */
	if (p_reduce_voltage != 0.0 || p_enhance_voltage != 0.0)
	{
		if (get_vid_with_offset(mastercpu->steps[mastercpu->step_index]->vid_orig, &vid_with_offset) != OK)
			return NOTOK;

		mastercpu->steps[mastercpu->step_index]->vid = vid_with_offset;
	}

	/* Build message for voltage info */
	if (msr_vid_to_voltage(mastercpu->steps[mastercpu->step_index]->vid, &voltage) != OK)
		return NOTOK;

	if (msr_vid_to_voltage(mastercpu->steps[mastercpu->step_index]->vid_orig, &voltage_orig) != OK)
		return NOTOK;

	libcpupowerd_message(LIBCPUPOWERD_MESSAGE_INFO, "Setting cpu undervolting voltage from %.4f V to %.4f V.\n", voltage_orig, voltage);

	/* Now under-/overvolting */
	if (set_vid(mastercpu->cpuid, mastercpu->steps[mastercpu->step_index]->vid) != OK)
		return NOTOK;

	libcpupowerd_message(LIBCPUPOWERD_MESSAGE_INFO, "Setting cpu undervolting voltage done.\n");

	libcpupowerd_get_timeofday_duration(&seconds, &useconds);
	libcpupowerd_message(LIBCPUPOWERD_MESSAGE_INFO, "Duration time is ");
	if (seconds > 0)
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_INFO, "%ld sec ", seconds);

	libcpupowerd_message(LIBCPUPOWERD_MESSAGE_INFO, "%ld usec.\n", useconds);

	return OK;
}


/* Check if we must change the frequency */
int decide_change(mastercpu_t *mastercpu, enum modes *mode)
{
	float load;
	float load_percent;

	if (cpu_get_load(mastercpu->cpuid, &load) != OK)
		return NOTOK;

	load_percent = 100 * load;

	libcpupowerd_message(LIBCPUPOWERD_MESSAGE_INFO, "Max load is %.2f percent on mastercpuid %d.\n", load_percent, mastercpu->cpuid);
	
	if (load_percent >= (float)p_speed_up)
		*mode = UPPER;
	else if (load_percent <= (float)p_speed_down)
		*mode = LOWER;
	else
		*mode = EQUAL;

	return OK;
}


/* Clean up */
int tini(void)
{
	int i, j;

	/* Clean up cpu interface */
	cpu_tini();
	/* Clean up msr interface */
	msr_tini();
	/* Clean up library */
	libcpupowerd_messages_set_default();

	/* Free the allocated memory */
	i = 0;
	while (mastercpus[i] != NULL)
	{
		j = 0;
		while (mastercpus[i]->steps[j] != NULL)
		{
			free(mastercpus[i]->steps[j]);
			j++;
		}
		
		free(mastercpus[i]->steps);
		free(mastercpus[i]);
		i++;
	}

	free(mastercpus);

	return OK;
}


/* Init all modules and needed structures */
int init(void)
{
	int i,j;
	int vendor;
	int family;
	int model;
	int fid_current;
	int vid_current;
	int program_work_mode;
	int frequencies_counter;
	int mastercpuids_counter;
	int quantity;
	int *cpuids;
	int *mastercpuids;
	int *frequencies;
	step_t *step;
	mastercpu_t *mastercpu;

	if (p_program_work_mode == 0)
		program_work_mode = LIBCPUPOWERD_MODE_FOREGROUND;
	else
		program_work_mode = LIBCPUPOWERD_MODE_DAEMON;

	/* Set message properties in cpupowerd library */
	if (libcpupowerd_message_set_properties(NAME, program_work_mode, p_message_level) != LIBCPUPOWERD_OK)
		return NOTOK;

	show_version();

	/* Init cpuid */
	if (cpuid_init() != OK)
		return NOTOK;

	if (cpuid_get_cpuids(&cpuids, &quantity) != OK)
		return NOTOK;

	if (cpuid_get_cpuinfo(&vendor, &family, &model) != OK)
		return NOTOK;

	/* Init cpu interface */
	if (cpu_init(cpuids) != OK)
		return NOTOK;

	/* Init msr interface */
	if (msr_init(vendor, family, model) != OK)
		return NOTOK;

	/* Init the cpu structure for the system */
	if (cpu_get_mastercpuids(&mastercpuids, &mastercpuids_counter) != OK)
		return NOTOK;

	/* Dynamic array from mastercpu struct */
	if ((mastercpus = (mastercpu_t **)malloc(sizeof(mastercpu_t *) * (mastercpuids_counter + 1))) == NULL)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Can't allocate memory for mastercpu array!\n");
		return NOTOK;
	}

	for (i = 0; i < mastercpuids_counter; i++)
	{
		/* Allocate memory for mastercpu struct */
		if ((mastercpu = (mastercpu_t *)malloc(sizeof(mastercpu_t))) == NULL)
		{
			libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Can't allocate memory for mastercpu struct!\n");
			return NOTOK;
		}

		/* Fill mastercpu struct with datas */
		memset(mastercpu, 0, sizeof(mastercpu_t));
		mastercpu->cpuid = mastercpuids[i];

		/* Get default governor */
		if (cpu_get_governor(mastercpu->cpuid, mastercpu->governor_default, (int)sizeof(mastercpu->governor_default)) != OK)
			return NOTOK;

		/* Set userspace governor if neccessary */
		if (strncmp(mastercpu->governor_default, GOVERNOR, strlen(GOVERNOR)) != 0)
		{
			if (cpu_set_governor(mastercpu->cpuid, GOVERNOR) != OK)
				return NOTOK;
		}

		/* Check msr access */
		if (msr_check_access(mastercpu->cpuid) != OK)
			return NOTOK;

		/* Get current vid and fid from cpu */
		if (msr_get_status(mastercpu->cpuid, &vid_current, &fid_current) != OK)
			return NOTOK;

		/* Fill step structs */
		if (cpu_get_frequencies(mastercpu->cpuid, &frequencies, &frequencies_counter) != OK)
			return NOTOK;

		/* Dynamic array from steps in mastercpu struct */
		if ((mastercpu->steps = (step_t **)malloc(sizeof(step_t *) * (frequencies_counter + 1))) == NULL)
		{
			libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Can't allocate memory for step array!\n");
			return NOTOK;
		}

		mastercpu->step_index = -1;
		for (j = 0; j < frequencies_counter; j++)
		{
			if ((step = (step_t *)malloc(sizeof(step_t))) == NULL)
			{
				libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Can't allocate memory for step struct!\n");
				return NOTOK;
			}

			memset(step, 0, sizeof(step_t));

			step->is_set_from_configfile = 0;

			if (msr_frequency_to_fid(frequencies[j], &step->fid) != OK)
				return NOTOK;

			if (step->fid == fid_current)
			{
				mastercpu->step_index = j;
				/* Set vid and vid_orig to the same value, because on init no undervolting is active */
				step->vid = vid_current;
				step->vid_orig = vid_current;
			}
			else
			{
				/* Vid must set over the config file */
				step->vid = -1;		 
				/* After changing frequency -> vid_orig is readed from the msr and stored */
				step->vid_orig = -1; 
			}

			mastercpu->steps[j] = step;
		}

		mastercpu->step_index_max = frequencies_counter - 1;
		if (mastercpu->step_index < 0)
		{
			libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "No current frequency in steps found!\n");
			return NOTOK;
		}

		mastercpu->steps[frequencies_counter] = NULL;
		free(frequencies);

		mastercpus[i] = mastercpu;
	}

	mastercpus[mastercpuids_counter] = NULL;
	free(mastercpuids);

	return OK;
}


/* Show the status */
int show_status(void)
{
	int i,j;
	int model;
	int family;
	int vendor;
	int frequency;
	int quantity_affectedcpuids;
	float voltage;
	int *affectedcpuids;

	if (cpuid_get_cpuinfo(&vendor, &family, &model) != OK)
		return NOTOK;

	printf("Vendor                      : ");
	switch (vendor)
	{
		case VENDOR_AMD:
			printf("AMD");
			break;
		case VENDOR_INTEL:
			printf("Intel");
			break;
		default:
			printf("Unknown");
	}

	printf("\nFamily                      : ");
	switch (family)
	{
		case AMD_K8:
			printf("K8");
			break;
		case AMD_K10:
			printf("K10");
			break;
		case INTEL_PENTIUM_M:
			printf("Pentium M");
			break;
		case INTEL_CORE:
			printf("Core");
			break;
		case INTEL_CORE_2:
			printf("Core2");
			break;
		default:
			printf("Unknown");
	}
	printf("\n");

	printf("Model                       : %d\n", model);

	i = 0;
	while (mastercpus[i] != NULL)
	{
		printf("  Mastercpuid               : %d\n", mastercpus[i]->cpuid);
		printf("    Affected cpuids         :");

		if (cpu_get_affectedcpuids(mastercpus[i]->cpuid, &affectedcpuids, &quantity_affectedcpuids) != OK)
			return NOTOK;

		j = 0;
		while (affectedcpuids[j] >= 0)
		{
			printf(" %d", affectedcpuids[j]);
			j++;
		}
		printf("\n");

		free(affectedcpuids);

		if (msr_vid_to_voltage(mastercpus[i]->steps[mastercpus[i]->step_index]->vid, &voltage) != OK)
			return NOTOK;

		if (msr_fid_to_frequency(mastercpus[i]->steps[mastercpus[i]->step_index]->fid, &frequency) != OK)
			return NOTOK;

		printf("    Current voltage (VID)   : %.4f V (%d)\n", voltage, mastercpus[i]->steps[mastercpus[i]->step_index]->vid);
		printf("    Current frequency (FID) : %d MHz (%d)\n", frequency, mastercpus[i]->steps[mastercpus[i]->step_index]->fid);

		printf("    Supported frequencies   :");
		j = 0;
		while (mastercpus[i]->steps[j] != NULL)
		{
			if (msr_fid_to_frequency(mastercpus[i]->steps[j]->fid, &frequency) != OK)
				return NOTOK;

			printf(" %d", frequency);
			j++;
		}
		printf(" MHz\n");

		i++;
	}

	return OK;
}


/* Show a message and exit the program */
void show_exitmessage(char *message)
{
	if (message != NULL)
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "%s\n", message);

	libcpupowerd_message(LIBCPUPOWERD_MESSAGE_INFO, "Try `%s --help' for more information.\n", NAME);

	exit(20);
}


/* Show version */
int show_version(void)
{
	libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "%s %d.%d.%d written by %s.\n", NAME, MAJOR, MINOR, UPDATE, AUTHOR);
	libcpupowerd_message(LIBCPUPOWERD_MESSAGE_WARNING, "WARNING: This program could cause damage to your Hardware!\n");

	return OK;
}


/* Show help */
int show_help(void)
{
	show_version();

	printf("\nExample: %s -a cpupowerd.conf\n", NAME);
	printf("Example: %s -c cpupowerd.conf -d -U 90\n\n", NAME);
	printf("Options:\n");
	printf("  -h, --help                print this help\n");
	printf("  -v, --version             print version and exit\n");
	printf("  -d, --daemonize           run program as a daemon\n");
	printf("  -m, --messagelevel ARG    level of the cpupowerd messages,\n");
	printf("                            ARG can be 0,1 or 2 (default = %d):\n", p_message_level);
	printf("                              0=ERROR, 1=WARNING, 2=INFO\n");
	printf("  -c, --config ARG          use parameters in ARG configuration file\n");
	printf("  -a, --autoconfig ARG      write automatic configuration parameters to file\n");
	printf("                            ARG\n");
	printf("  -t, --time ARG            ARG is the time in milliseconds the program waits\n");
	printf("                            to check the load of the mastercpus (default\n");
	printf("                            = %d)\n", LOOP_TIME);
	printf("  -s, --status              show status of the mastercpus\n");
	printf("  -l, --laptop              activate laptop mode check\n");
	printf("  -r, --reducevoltage ARG   reduce voltage for all cpus and frequencies at\n");
	printf("                            ARG, expressed as a float value. The highest value\n");
	printf("                            for reduce is %.4f V\n", REDUCE_VOLTAGE_MAX);
	printf("  -e, --enhancevoltage ARG  enhance voltage for all cpus and frequencies at\n");
	printf("                            ARG, expressed as a float value. The highest value\n");
	printf("                            for enhance is %.4f V\n", ENHANCE_VOLTAGE_MAX);
	printf("  -S, --strategy ARG        strategy of operation, ARG can be 0,1,2,3,4,5 or 6\n");
	printf("                            (default=%d):\n", (int)p_strategy);
	printf("                              0=MINIMUM, 1=SINE, 2=AGGRESSIVE, 3=PASSIVE,\n");
	printf("                              4=BSD, 5=LEAPS, 6=MAXIMUM\n");
	printf("  -U, --up ARG              set ARG to be the default percentage of use causing\n");
	printf("                            increase the cpu frequency [0 .. 100, default %d]\n", SPEED_UP_INIT);
	printf("  -D, --down ARG            set ARG to be the default percentage of use causing\n");
	printf("                            decrease the cpu frequency [0 .. 100, default %d]\n", SPEED_DOWN_INIT);
	printf("  -V, --voltage ARG1|ARG2   set ARG2 in volts as the new voltage, expressed as\n");
	printf("                            a float value for mastercpuid ARG1\n");
	printf("  -F, --freq ARG1|ARG2      set ARG2 in MHz as the new frequency, expressed as\n");
	printf("                            a int value for mastercpuid ARG1\n\n");

	printf("Report bugs to %s <%s>\n", AUTHOR, EMAIL);

	return OK;
}


/* Parse program options */
int parse_options(int argc, char **argv)
{
	int c;
	char *pStr;
	char *pStr2;
	static struct option longopts[] = {
		{ "help",	  		no_argument, 		NULL, 'h'},
		{ "version",		no_argument, 		NULL, 'v'},
		{ "daemonize",		no_argument, 		NULL, 'd'},
		{ "foreground",		no_argument, 		NULL, 'f'},
		{ "status",			no_argument, 		NULL, 's'},
		{ "laptop",			no_argument, 		NULL, 'l'},
		{ "messagelevel",	required_argument, 	NULL, 'm'},
		{ "reducevoltage",	required_argument, 	NULL, 'r'},
		{ "enancvoltage",	required_argument, 	NULL, 'e'},
		{ "strategy",		required_argument, 	NULL, 'S'},
		{ "autoconfig",		required_argument, 	NULL, 'a'},
		{ "config",			required_argument, 	NULL, 'c'},
		{ "up",				required_argument, 	NULL, 'U'},
		{ "down",			required_argument, 	NULL, 'D'},
		{ "voltage",		required_argument, 	NULL, 'V'},
		{ "frequency",		required_argument, 	NULL, 'F'},
		{ "time",			required_argument, 	NULL, 't'},
		{ NULL,				0, 					NULL, 0}
	};

	/* Parse command line args */
	while ((c = getopt_long(argc, argv, "hvdfslm:r:e:S:a:c:U:D:V:F:t:", longopts, 0)) != EOF)
	{
		switch(c) 
		{
			case 'd':  /* --daemonize */
				p_program_work_mode = 1;
				break;
			case 'f':  /* --foreground */
				libcpupowerd_message(LIBCPUPOWERD_MESSAGE_INFO, "Option '-f' is deprecated and has no effect anymore!\n");
				break;
			case 'h':  /* --help */
				show_help();
				exit(0);
			case 'v':  /* --version */
				show_version();
				exit(0);
			case 's':  /* --status */
				p_status = 1;
			case 'l':  /* --laptop */
				p_laptop = 1;
				break;
			case 'm':  /* --tracelevel <level> */
				p_message_level = atof(optarg);
				if (p_message_level > LIBCPUPOWERD_MESSAGE_INFO || p_message_level < LIBCPUPOWERD_MESSAGE_ERROR)
					show_exitmessage("Invalid message level specified!");
				break;
			case 'r':  /* --reducevoltage <voltage> */
				p_reduce_voltage = atof(optarg);
				if (p_reduce_voltage > REDUCE_VOLTAGE_MAX + 0.000001)
					show_exitmessage("Reduced voltage value too high - only possible over the config file!");
				break;
			case 'e':  /* --enhancevoltage <voltage> */
				p_enhance_voltage = atof(optarg);
				if (p_enhance_voltage > ENHANCE_VOLTAGE_MAX + 0.000001)
					show_exitmessage("Enhanced voltage value too high - only possible over the config file!");
				break;
			case 'S':  /* --strategy */
				p_strategy = atoi(optarg);
				if ((p_strategy < 0) || (p_strategy > 6)) 
					show_exitmessage("Invalid strategy specified!");
				break;			
			case 'a':  /* --autoconfig <autoconfig-file> */
				p_autoconfigfile = optarg;
				break;
			case 'c':  /* --config <config-file> */
				p_configfile = optarg;
				break;
			case 'U':  /* --up <percentage> */
				p_speed_up = atoi(optarg);
				if (p_speed_up < 0 || p_speed_up > 100 )
					show_exitmessage("--up <percentage>: only values between 0 and 100 are allowed!");
				break;
			case 'D':  /* --down <percentage> */
				p_speed_down = atoi(optarg);
				if (p_speed_down < 0 || p_speed_down > 100 )
					show_exitmessage("--down <percentage>: only values between 0 and 100 are allowed!");
				break;
			case 'V':  /* --voltage <cpu|voltage> */
				pStr = strdup(optarg);
				pStr = strtok(pStr, "|");
				pStr2 = strtok(NULL, "|");
				if (pStr != NULL)
					p_voltage_cpuid = atoi(pStr);
				if (pStr2 != NULL)
					p_voltage = atof(pStr2);
				if (pStr == NULL || pStr2 == NULL)
					show_exitmessage("Please specify mastercpuid and voltage!");
				break;
			case 'F':  /* --frequency <cpu|frequency> */
				pStr = strdup(optarg);
				pStr = strtok(pStr, "|");
				pStr2 = strtok(NULL, "|");
				if (pStr != NULL)
					p_frequency_cpuid = atoi(pStr);
				if (pStr2 != NULL)
					p_frequency = atol(pStr2);
				if (pStr == NULL || pStr2 == NULL)
					show_exitmessage("Please specify mastercpuid and frequency!");
				break;
			case 't':  /* --time <msec> */
				p_time = atoi(optarg);
				break;
			default:
				show_exitmessage(NULL);
				return OK;
		}
	}

	/* Additional checks */
	if (p_speed_up < p_speed_down)
		show_exitmessage("--up and --down: Up threshold can't be lower than down threshold!");

	if (p_reduce_voltage != 0.0 && p_enhance_voltage != 0.0)
		show_exitmessage("Only one of the options --reducevoltage or --enhancevoltage can be used!");

	if (p_reduce_voltage != 0.0 && p_enhance_voltage != 0.0)
		show_exitmessage("You have specified a reduced and enhanced voltage. This isn't possible!");

	if (p_program_work_mode != 0 && (p_configfile == NULL && p_reduce_voltage == 0.0 && p_enhance_voltage == 0.0))
		show_exitmessage("Please specify --config ARG, --reducevoltage or --enhancevoltage!");

	if (p_reduce_voltage != 0.0 && p_configfile != NULL)
	{
		p_configfile = NULL;
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_INFO, "You have specified a reduced voltage and a config file. The reduced voltage will be used!\n");
	}

	if (p_enhance_voltage != 0.0 && p_configfile != NULL)
	{
		p_configfile = NULL;
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_INFO, "You have specified an enhanced voltage and a config file. The enhanced voltage will be used!\n");
	}

	return OK;
}


/* Clean up and exit */
void terminate(int signum)
{
	int i;

	i = 0;
	while (mastercpus[i] != NULL)
	{
		/* Restore original voltage */
		if (set_vid(mastercpus[i]->cpuid, mastercpus[i]->steps[mastercpus[i]->step_index]->vid_orig) != OK)
			libcpupowerd_message(LIBCPUPOWERD_MESSAGE_WARNING, "Can't restore original voltage!\n");

		/* Restore original governor */
		if (cpu_set_governor(mastercpus[i]->cpuid, mastercpus[i]->governor_default) != OK)
				libcpupowerd_message(LIBCPUPOWERD_MESSAGE_WARNING, "Can't restore governor!\n");

		i++;
	}

	if (tini() != OK)
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_WARNING, "Tini failed!\n");

	libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "%s exiting.\n", NAME);

	exit(0);
}


/* main */
int main (int argc, char **argv)
{
	int i;
	int vid_tmp;
	int fid_tmp;
	int vid_with_offset;
	enum modes mode;

	parse_options(argc, argv);

	if (getuid() != 0)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "You must be root to run %s!\n", NAME);
		exit(1);
	}

	if (init() != OK)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Initialisation failed!\n");
		exit(2);
	}

	if (p_configfile != NULL)
	{
		if (read_config(p_configfile) != OK)
		{
			libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Couldn't read config file correctly!\n");
			exit(3);
		}
	}

	if (p_status != 0)
	{
		show_status();
		exit(4);
	}

	if (p_frequency_cpuid >= 0 && p_frequency > 0)
	{
		if (msr_frequency_to_fid(p_frequency, &fid_tmp) != OK)
			exit(4);

		if (set_fid(p_frequency_cpuid, fid_tmp) != OK)
			exit(5);
	}

	if (p_voltage_cpuid >= 0 && p_voltage > 0.0)
	{
		if (msr_voltage_to_vid(p_voltage, &vid_tmp) != OK)
			exit(6);

		if (set_vid(p_voltage_cpuid, vid_tmp) != OK)
			exit(7);
	}

	if (p_autoconfigfile != NULL)
	{
		if (write_config(p_autoconfigfile) != OK)
			exit(8);
	}

	/* Make the program run as a daemon */
	if (p_program_work_mode != 0)
	{
		if (daemon(0, 0))
		{
			libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Couldn't create daemon!\n");
			exit(9);
		}
    }

	if (p_configfile != NULL || p_reduce_voltage != 0.0 || p_enhance_voltage != 0.0)
	{
		/* Lets set up a exit handler */
		signal(SIGTERM, terminate);
		signal(SIGINT, terminate);

		/* First undervolting */
		i = 0;
		while (mastercpus[i] != NULL)
		{
			if (p_reduce_voltage != 0.0 || p_enhance_voltage != 0.0)
			{
				if (get_vid_with_offset(mastercpus[i]->steps[mastercpus[i]->step_index]->vid_orig,
                               &vid_with_offset) != OK)
				{
					libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Can't calculate vid with offset!\n");
					exit(10);
				}

				mastercpus[i]->steps[mastercpus[i]->step_index]->vid = vid_with_offset;
			}

			if (set_vid(mastercpus[i]->cpuid, mastercpus[i]->steps[mastercpus[i]->step_index]->vid) != OK)
			{
				libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "First undervolting failed!\n");
				exit(11);
			}
		
			i++;
		}

		/* loop */
		while (1)
		{
			usleep(p_time * 1000);
			i = 0;
			while (mastercpus[i] != NULL)
			{
				/* Check change */
				if (decide_change(mastercpus[i], &mode) != OK)
				{
					libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Decide change failed!\n");
					exit(12);
				}
				else
				{
					if (mode != EQUAL)
					{
						/* Change frequency and voltage */
						if (change_cpustep(mastercpus[i], mode) != OK)
						{
							libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Change cpu step failed!\n");
							exit(13);
						}
					}
				}

				i++;
			}
		}
	}
	else
		tini();

	return 0;
}


