/************************************************************************/
/* Copyright (C) 2008-2009 Markus Strobl <markus.strobl@gmx.at>         */
/*                                                                      */
/* cpupowerd is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by    */
/* the Free Software Foundation; either version 2 of the License.       */
/*                                                                      */
/* cpupowerd is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of           */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */
/*                                                                      */
/* You should have received a copy of the GNU General Public License    */
/* along with GNU Emacs; see the file COPYING.  If not, write to        */
/* the Free Software Foundation, Inc., 59 Temple Place - Suite 330,     */
/* Boston, MA 02111-1307, USA.                                          */
/************************************************************************/
/* $Id: libcpupowerd.c 369 2009-01-01 13:05:07Z markus $                */
/************************************************************************/


/* Includes */
#include <syslog.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/sysctl.h>
#ifdef __FreeBSD__
#include <sys/param.h>
#if __FreeBSD_version < 800042
#include <cpu.h>
#define RDMSR			CPU_RDMSR
#define WRMSR			CPU_WRMSR
#define CPUID			CPU_CPUID
#define INTERFACEPATH   "/dev/cpu"
#else
#include <sys/cpuctl.h>
#define RDMSR			CPUCTL_RDMSR
#define WRMSR			CPUCTL_WRMSR
#define CPUID			CPUCTL_CPUID
#define INTERFACEPATH  	"/dev/cpuctl"
#endif
#endif

#include "libcpupowerd.h"


/* Global variables */
static int work_mode = LIBCPUPOWERD_MODE_FOREGROUND;		/* Default worke mode is foreground -> printf */
static int message_typ_limit = LIBCPUPOWERD_MESSAGE_INFO;	/* Default message level */
static struct timeval begin_timeofday;						/* Struct for duration time calculation */


/* Prototypes */
static int libcpupowerd_get_interfacefile(int cpuid, char *file, size_t size, int type);
#ifdef __FreeBSD__
static int libcpupowerd_read_msr_ioctl(char *msrfile, unsigned long index, unsigned long long *msrvalue);
static int libcpupowerd_write_msr_ioctl(char *msrfile, unsigned long index, unsigned long long *msrvalue);
static int libcpupowerd_read_cpuid_ioctl(char *cpuidfile, unsigned long index, unsigned long *eax, unsigned long *ebx, unsigned long *ecx, unsigned long *edx);
#endif


/* Functions */
/* Set messages properties */
int libcpupowerd_message_set_properties(char* program_name, int program_work_mode, int program_message_typ_limit)
{
	if (program_work_mode != LIBCPUPOWERD_MODE_DAEMON && program_work_mode != LIBCPUPOWERD_MODE_FOREGROUND)
		return LIBCPUPOWERD_NOTOK;

	if (work_mode == LIBCPUPOWERD_MODE_DAEMON && program_work_mode == LIBCPUPOWERD_MODE_FOREGROUND)
		closelog();

	if (work_mode == LIBCPUPOWERD_MODE_FOREGROUND && program_work_mode == LIBCPUPOWERD_MODE_DAEMON)
		openlog(program_name, LOG_PID, LOG_DAEMON);

	work_mode = program_work_mode;
	message_typ_limit = program_message_typ_limit;

	return LIBCPUPOWERD_OK;
}


/* Set messages default properties */
int libcpupowerd_messages_set_default(void)
{
	if (work_mode == LIBCPUPOWERD_MODE_DAEMON)
		closelog();

	work_mode = LIBCPUPOWERD_MODE_FOREGROUND;
	message_typ_limit = LIBCPUPOWERD_MESSAGE_INFO;

	return LIBCPUPOWERD_OK;
}


/* Logging messages */
int libcpupowerd_message(int message_typ, char *message_format, ...)
{
	if (message_typ <= message_typ_limit)
	{
		va_list ap;
    	va_start (ap, message_format);

		if (work_mode == LIBCPUPOWERD_MODE_DAEMON)
		{
			switch (message_typ)
			{
				case LIBCPUPOWERD_MESSAGE_INFO:
					vsyslog(LOG_INFO, message_format, ap);
					break;
				case LIBCPUPOWERD_MESSAGE_WARNING:
					vsyslog(LOG_WARNING, message_format, ap);
					break;
				case LIBCPUPOWERD_MESSAGE_ERROR:
					vsyslog(LOG_ERR, message_format, ap);
					break;
				default:
					return LIBCPUPOWERD_NOTOK;
			}
		}
		else
			vprintf(message_format, ap);
	}

	return LIBCPUPOWERD_OK;
}


/* Read from msr file */
int libcpupowerd_read_msr(char *msrfile, unsigned long seek_position, unsigned long *low, unsigned long *high)
{
	unsigned long long msrvalue;

	#ifdef __linux__
	int readed_bytes;

	if (libcpupowerd_read_file(msrfile, seek_position, (int)sizeof(msrvalue), (void *)&msrvalue, (int)sizeof(msrvalue), &readed_bytes) != LIBCPUPOWERD_OK)
		return LIBCPUPOWERD_NOTOK;
	#elif __FreeBSD__
	if (libcpupowerd_read_msr_ioctl(msrfile, seek_position, &msrvalue) != LIBCPUPOWERD_OK)
		return LIBCPUPOWERD_NOTOK;
	#else
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Read msr not implemented!\n");
		return LIBCPUPOWERD_NOTOK;
	#endif

	*high = msrvalue >> 32;
	*low = msrvalue;

	return LIBCPUPOWERD_OK;
}


/* Write to msr file */
int libcpupowerd_write_msr(char *msrfile, unsigned long seek_position, unsigned long long *msrvalue)
{
	#ifdef __linux__
	int write_bytes;

	write_bytes = 8;

	if (libcpupowerd_write_file(msrfile, seek_position, write_bytes, (void *)msrvalue) != LIBCPUPOWERD_OK)
		return LIBCPUPOWERD_NOTOK;

	#elif __FreeBSD__
	if (libcpupowerd_write_msr_ioctl(msrfile, seek_position, msrvalue) != LIBCPUPOWERD_OK)
		return LIBCPUPOWERD_NOTOK;
	#else
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Write msr not implemented!\n");
		return LIBCPUPOWERD_NOTOK;
	#endif

	return LIBCPUPOWERD_OK;
}


/* Read from cpuid file */
int libcpupowerd_read_cpuid(char *cpuidfile, unsigned long seek_position, unsigned long *eax, unsigned long *ebx, unsigned long *ecx, unsigned long *edx)
{
	#ifdef __linux__
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Read cpuid not implemented!\n");
		return LIBCPUPOWERD_NOTOK;
	#elif __FreeBSD__
	if (libcpupowerd_read_cpuid_ioctl(cpuidfile, seek_position, eax, ebx, ecx, edx) != LIBCPUPOWERD_OK)
		return LIBCPUPOWERD_NOTOK;
	#else
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Read cpuid not implemented!\n");
		return LIBCPUPOWERD_NOTOK;
	#endif

	return LIBCPUPOWERD_OK;
}


/* Read from file */
int libcpupowerd_read_file(char *file, unsigned long seek_position, int bytes, void *buffer, int buffer_size, int *readed_bytes)
{
	int fd;
	int result;
	int read_bytes;

	if (bytes != 0)
	{
		if (bytes > buffer_size)
		{
			libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Buffer size with %d is to small for reading %d bytes!\n", buffer_size, bytes);
			return LIBCPUPOWERD_NOTOK;
		}
		read_bytes = bytes;
	}
	else
		read_bytes = buffer_size;

	if ((fd = open(file, O_RDONLY)) == -1)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Opening file %s for reading failed!\n", file);
		return LIBCPUPOWERD_NOTOK;
	}

	if (lseek(fd, (off_t)seek_position, SEEK_SET) == -1)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Seeking in file %s failed!\n", file);
		return LIBCPUPOWERD_NOTOK;
	}

	result = (int)read(fd, buffer, (size_t)read_bytes);
	if (result == -1)
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Reading file %s failed!\n", file);

	if (close(fd) == -1)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Closeing file %s failed!\n", file);
		return LIBCPUPOWERD_NOTOK;
	}

	if (result == -1)
		return LIBCPUPOWERD_NOTOK;

	*readed_bytes = result;

	if (bytes != 0)
	{
		if (*readed_bytes != read_bytes)
		{
			libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "The count of readed bytes should be %d, but is only %d!\n", read_bytes, readed_bytes);
			return LIBCPUPOWERD_NOTOK;		
		}
	}

	return LIBCPUPOWERD_OK;
}


/* Read from file with alloc memory */
int libcpupowerd_read_file_dynamic(char *file, char **buffer, int *size)
{
	int fd;
	int result;
	int counter;
	int buffer_size;
	char *buffer_tmp;
	char tmpbuffer[128];

	buffer_size = 512;
	if ((buffer_tmp = malloc(buffer_size * sizeof(char))) == NULL)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Allocating memory failed!\n");
		return LIBCPUPOWERD_NOTOK;
	}

	if ((fd = open(file, O_RDONLY)) == -1)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Opening failed for file %s!\n", file);
		return LIBCPUPOWERD_NOTOK;
	}

	if (lseek(fd, 0, SEEK_SET) == -1)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Seeking failed in file %s!\n", file);
		return LIBCPUPOWERD_NOTOK;
	}

	counter = 0;
	buffer_tmp[counter] = '\0';
	while ((result = (int)read(fd, &tmpbuffer, sizeof(tmpbuffer) - 1)) > 0)
	{
		tmpbuffer[result] = '\0';
		if (counter + result >= buffer_size - 1 )
		{
			buffer_size *= 2;
			if ((buffer_tmp = realloc(buffer_tmp, buffer_size)) == NULL)
			{
				libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Reallocating memory failed!\n");
				return LIBCPUPOWERD_NOTOK;
			}
		}
	 	strncat(buffer_tmp, tmpbuffer, result);
		counter += result;
	}

	if (result == -1)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Reading failed in file %s!\n", file);
		return LIBCPUPOWERD_NOTOK;
	}

	buffer_tmp[counter] = '\0';

	if (close(fd) == -1)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Closeing failed for file %s!\n", file);
		return LIBCPUPOWERD_NOTOK;
	}

	*buffer = buffer_tmp;
	*size = buffer_size;

	return LIBCPUPOWERD_OK;
}


/* Write to file */
int libcpupowerd_write_file(char *file, unsigned long seek_position, int write_bytes, void *buffer)
{
	int fd;
	int result;

	if ((fd = open(file, O_WRONLY)) == -1)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Opening file %s for writing failed!\n", file);
		return LIBCPUPOWERD_NOTOK;
	}

	if (lseek(fd, (off_t)seek_position, SEEK_SET) == -1)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Seeking in file %s failed!\n", file);
		return LIBCPUPOWERD_NOTOK;
	}

	result = (int)write(fd, buffer, (size_t)write_bytes);
	if (result == -1)
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Writing to file %s failed!\n", file);

	if (close(fd) == -1)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Closeing file %s failed!\n", file);
		return LIBCPUPOWERD_NOTOK;
	}

	if (result == -1)
		return LIBCPUPOWERD_NOTOK;

	if (result != write_bytes)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "The count of writed bytes should be %d, but is only %d!\n", write_bytes, result);
		return LIBCPUPOWERD_NOTOK;
	}

	return LIBCPUPOWERD_OK;
}


/* Busy waiting, no better solution found :-(, nanosleep operate like the same up to 1 ms */
int libcpupowerd_usleep(int sleep_usec)
{
	struct timeval start;
	struct timeval stop;
	double sleeptime;

	if (sleep_usec > 100)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Busy waiting doesn't make sense over 100 useconds!\n");
		return LIBCPUPOWERD_NOTOK;
	}

	if (gettimeofday(&start, NULL) == -1)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_WARNING, "Start timeofday can't be set!\n");
		return LIBCPUPOWERD_NOTOK;
	}

	do
	{
		if (gettimeofday(&stop, NULL) == -1)
		{
			libcpupowerd_message(LIBCPUPOWERD_MESSAGE_WARNING, "Stop timeofday can't be set!\n");
			return LIBCPUPOWERD_NOTOK;
		}

		sleeptime = ((stop.tv_sec + (stop.tv_usec / 1000000.) - (start.tv_sec + (start.tv_usec / 1000000.))) * 1000000.);
	}
	while (sleeptime < sleep_usec);

	return LIBCPUPOWERD_OK;
}


/* Set timeofday */
int libcpupowerd_set_timeofday(void)
{
	if (gettimeofday(&begin_timeofday, NULL) == -1)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_INFO, "Begin timeofday can't be set!\n");
		return LIBCPUPOWERD_NOTOK;
	}

	return LIBCPUPOWERD_OK;
}


/* Get timeofday duration */
int libcpupowerd_get_timeofday_duration(long *seconds, long *useconds)
{
  	long seconds_tmp = 0;
	long useconds_tmp = 0;
	struct timeval end_timeofday;

	if (gettimeofday(&end_timeofday, NULL) == -1)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_INFO, "End timeofday can't be get!\n");
		return LIBCPUPOWERD_NOTOK;		
	}

	seconds_tmp = end_timeofday.tv_sec - begin_timeofday.tv_sec;
	useconds_tmp = end_timeofday.tv_usec - begin_timeofday.tv_usec;
	if (useconds_tmp < 0)
	{
		useconds_tmp += 1000000;
		seconds_tmp--;
	}

	*seconds = seconds_tmp;
	*useconds = useconds_tmp;

	return LIBCPUPOWERD_OK;
}


/* Get integers from string */
int libcpupowerd_get_integers_from_string(char *string, char *parse_format, char delimiter, int **integers, int *quantity)
{
	int i;
	char *p, *q;

	/* isspace(*p) && *p != ' ' : '\f' && '\n' && '\r' && '\t' && '\v' */
	if (isdigit(delimiter) || (isspace(delimiter) && delimiter != ' '))
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Delimiter |%c| can't be used to get integers from string |%s|!\n", delimiter, string);
		return LIBCPUPOWERD_NOTOK;
	}

	/* Format string and remove unneeded characters */
	q = p = string;
	while (*p == delimiter)
		p++;

	for (; *p != '\0'; p++)
	{
		if (*p == delimiter && (*(p + 1) == delimiter || *(p + 1) == '\0' || (isspace(*(p + 1)) && *(p + 1) != ' ')))
			continue;

		if (isspace(*p) && *p != ' ')
			continue;

		if (*p != delimiter && !isdigit(*p))
		{
			/* Only on FreeBSD '/' is allowed */
			if (strchr(parse_format, '/') != NULL)
			{
				if (*p != '/')
				{
					libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "String |%s| contain the not allowed character |%c|!\n", string, *p);
					return LIBCPUPOWERD_NOTOK;
				}
			}
			else
			{
				libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "String |%s| contain the not allowed character |%c|!\n", string, *p);
				return LIBCPUPOWERD_NOTOK;
			}
		}

		*q = *p;
		q++;
	}
	*q = '\0';

	if (strlen(string) < 1)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "String |%s| doesn't contain integers!\n", string);
		return LIBCPUPOWERD_NOTOK;
	}

	/* Get quantity of intergers */
	*quantity = 1;
	p = string;
	while ((q = strchr(p, delimiter)) != NULL)
	{
		p = q + 1;
		(*quantity)++;
	}

	/* Malloc needed memory */
	if ((*integers = malloc((*quantity + 1) * sizeof(int))) == NULL)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Can't allocate memory for integers!\n");
		return LIBCPUPOWERD_NOTOK;
	}

	/* Get the integers */
	for (i = 0, p = string; i < *quantity; i++)
	{
		if ((q = strchr(p, delimiter)) != NULL)
			*q = '\0';

		if (sscanf(p, parse_format, &(*integers)[i]) != 1)
		{
			libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Can't get integer from string |%s|!\n", p);
			return LIBCPUPOWERD_NOTOK;
		}

		p = q + 1;
	}
	(*integers)[i] = -1;

	return LIBCPUPOWERD_OK;
}


/* Get timestamp */
int libcpupowerd_get_timestamp(unsigned long *timestamp)
{
	time_t timestamp_tmp;

	if ((timestamp_tmp = time(NULL)) == -1)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Create timestamp failed!\n");
		return LIBCPUPOWERD_NOTOK;
	}

	*timestamp = (unsigned long)timestamp_tmp;

	return LIBCPUPOWERD_OK;
}


/* Compare function for qsort */
int libcpupowerd_compare_ints(const void *a, const void *b)
{
	const int *la = (const int *) a;
	const int *lb = (const int *) b;

	return (*la > *lb) - (*la < *lb);
}


/* Get msr file */
int libcpupowerd_get_msrfile(int cpuid, char *file, size_t size)
{
	return libcpupowerd_get_interfacefile(cpuid, file, size, 0);
}


/* Get cpuid file */
int libcpupowerd_get_cpuidfile(int cpuid, char *file, size_t size)
{
	return libcpupowerd_get_interfacefile(cpuid, file, size, 1);
}


/* Get sysctl */
int libcpupowerd_getsysctl(char *name, void *ptr, size_t size)
{	
	#ifdef __linux__
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Getsysctl not implemented!\n");
		return LIBCPUPOWERD_NOTOK;
	#elif __FreeBSD__
	if (sysctlbyname(name, ptr, &size, NULL, 0) == -1)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Error by reading sysctl %s!\n", name);
		return LIBCPUPOWERD_NOTOK;
	}
	#else
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Getsysctl not implemented!\n");
		return LIBCPUPOWERD_NOTOK;
	#endif

	return LIBCPUPOWERD_OK;
}


static int libcpupowerd_get_interfacefile(int cpuid, char *file, size_t size, int type)
{
	if (size < 16)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Buffer for interface file name to small!\n");
		return LIBCPUPOWERD_NOTOK;
	}

	#ifdef __linux__
	if (type == 0)
		sprintf(file, "/dev/cpu/%d/msr", cpuid);
	else
		sprintf(file, "/dev/cpu/%d/cpuid", cpuid);

	return LIBCPUPOWERD_OK;
	#elif __FreeBSD__
	sprintf(file, "%s%d", INTERFACEPATH, cpuid);

	return LIBCPUPOWERD_OK;
	#else
	libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Path to kernel interface unknown!\n");
	return LIBCPUPOWERD_NOTOK;
	#endif
}


#ifdef __FreeBSD__
/* Read msr with ioctl */
static int libcpupowerd_read_msr_ioctl(char *msrfile, unsigned long index, unsigned long long *msrvalue)
{
  	int fd;
	#if __FreeBSD_version < 800042
 	cpu_msr_args_t args;
	#else
	cpuctl_msr_args_t args;
	#endif

	args.msr = index;

	if ((fd = open(msrfile, O_RDONLY)) == -1)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Opening msr file %s for reading failed!\n", msrfile);
		return LIBCPUPOWERD_NOTOK;
	}
	
	if (ioctl(fd, RDMSR, &args) == -1)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Reading msr file %s failed!\n", msrfile);
		return LIBCPUPOWERD_NOTOK;
	}

	if (close(fd) == -1)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Closeing msr file %s failed!\n", msrfile);
		return LIBCPUPOWERD_NOTOK;
	}

	*msrvalue = args.data;
 	
 	return LIBCPUPOWERD_OK;
}


/* Write msr with ioctl */
static int libcpupowerd_write_msr_ioctl(char *msrfile, unsigned long index, unsigned long long *msrvalue)
{
	int fd;
	#if __FreeBSD_version < 800042
 	cpu_msr_args_t args;
	#else
	cpuctl_msr_args_t args;
	#endif

	args.msr = index;
	args.data = *msrvalue;
	
	if ((fd = open(msrfile, O_WRONLY, 0)) == -1)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Opening msr file %s for writing failed!\n", msrfile);
		return LIBCPUPOWERD_NOTOK;
	}
	
	if (ioctl(fd, WRMSR, &args) == -1)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Writing msr file %s failed!\n", msrfile);
		return LIBCPUPOWERD_NOTOK;
	}

	if (close(fd) == -1)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Closeing msr file %s failed!\n", msrfile);
		return LIBCPUPOWERD_NOTOK;
	}

	return LIBCPUPOWERD_OK;
}


/* Read cpuid with ioctl */
static int libcpupowerd_read_cpuid_ioctl(char *cpuidfile, unsigned long index, unsigned long *eax, unsigned long *ebx, unsigned long *ecx, unsigned long *edx)
{
  	int fd;
	#if __FreeBSD_version < 800042
 	cpu_cpuid_args_t args;
	#else
	cpuctl_cpuid_args_t args;
	#endif

	args.level = index;

	if ((fd = open(cpuidfile, O_RDONLY)) == -1)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Opening cpuid file %s for reading failed!\n", cpuidfile);
		return LIBCPUPOWERD_NOTOK;
	}
	
	if (ioctl(fd, CPUID, &args) == -1)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Reading cpuid file %s failed!\n", cpuidfile);
		return LIBCPUPOWERD_NOTOK;
	}

	if (close(fd) == -1)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Closeing cpuid file %s failed!\n", cpuidfile);
		return LIBCPUPOWERD_NOTOK;
	}

	*eax = args.data[0];
	*ebx = args.data[1];
	*ecx = args.data[2];
	*edx = args.data[3];
 	
 	return LIBCPUPOWERD_OK;
}
#endif






