/************************************************************************/
/* Copyright (C) 2008-2009 Markus Strobl <markus.strobl@gmx.at>         */
/*                                                                      */
/* cpupowerd is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by    */
/* the Free Software Foundation; either version 2 of the License.       */
/*                                                                      */
/* cpupowerd is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of           */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */
/*                                                                      */
/* You should have received a copy of the GNU General Public License    */
/* along with GNU Emacs; see the file COPYING.  If not, write to        */
/* the Free Software Foundation, Inc., 59 Temple Place - Suite 330,     */
/* Boston, MA 02111-1307, USA.                                          */
/************************************************************************/
/* $Id: msr.c 369 2009-01-01 13:05:07Z markus $                         */
/************************************************************************/


/* Includes */
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>

#include "msr.h"
#include "cpupowerd.h"
#include "libcpupowerd.h"
#include "msr_amd.h"


/* Function pointers */
int (*ptr_msr_vid_to_voltage) (int, int, int, float *);			/* msr_vid_to_voltage */
int (*ptr_msr_voltage_to_vid) (int, int, float, int *);			/* msr_vid_to_voltage */
int (*ptr_msr_fid_to_frequency) (int, int, int, int *);			/* msr_fid_to_frequency */
int (*ptr_msr_frequency_to_fid) (int, int, int, int *);			/* msr_frequency_to_fid */
int (*ptr_msr_get_status) (int, int, char *, int *, int *);		/* msr_get_status */
int (*ptr_msr_set_status) (int, int, char *, int, int);			/* msr_set_voltage */
int (*ptr_msr_tini) (void);										/* msr_tini */


/* Global variables */
static int family;
static int model;


/* Functions */
/* Init */
int msr_init(int vendor, int cpu_family, int cpu_model)
{
	int result;

	family = cpu_family;
	model = cpu_model;

	switch (vendor)
	{
		case VENDOR_AMD:
			result = msr_amd_init(&ptr_msr_vid_to_voltage, &ptr_msr_voltage_to_vid,
								  &ptr_msr_fid_to_frequency, &ptr_msr_frequency_to_fid,
								  &ptr_msr_get_status, &ptr_msr_set_status,
								  &ptr_msr_tini);
			break;
		/*case VENDOR_INTEL:
			result = msr_intel_init(&ptr_msr_vid_to_voltage, &ptr_msr_voltage_to_vid, 
								    &ptr_msr_fid_to_frequency, &ptr_msr_frequency_to_fid, 
								    &ptr_msr_get_status, &ptr_msr_set_status, 
								    &ptr_msr_tini);
			break;*/
		default:
			libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Your cpu vendor isn't yet supported!\n");
			result = NOTOK;
	}

	if (result != OK)
		return NOTOK;

	return OK;
}


/* Clean up */
int msr_tini(void)
{
	return (*ptr_msr_tini)();
}


/* Check access to msr file */
int msr_check_access(int cpuid)
{
	int result;
	char msrfile[32];

	if (libcpupowerd_get_msrfile(cpuid, msrfile, sizeof(msrfile)) != LIBCPUPOWERD_OK)
		return NOTOK;

	if ((result = access(msrfile, R_OK|W_OK)) != 0)
	{
		if (errno == ENOENT)
			libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "MSR-File <%s> doesn't exist, please load the msr kernel module!\n", msrfile);

		if (errno == EACCES)
			libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "No read/write permission to MSR-File <%s>!\n", msrfile);

		if (errno == EROFS)
			libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "No write permission for the read-only MSR-File <%s>!\n", msrfile);

		return NOTOK;
	}

	return OK;
}


/* Calculates frequency from fid */
int msr_fid_to_frequency(int fid, int *frequency)
{
	return (*ptr_msr_fid_to_frequency)(family, model, fid, frequency);
}


/* Calculates fid from frequency */
int msr_frequency_to_fid(int frequency, int *fid)
{
	return (*ptr_msr_frequency_to_fid)(family, model, frequency, fid);
}


/* Calculates voltage from vid */
int msr_vid_to_voltage(int vid, float *voltage)
{
	return (*ptr_msr_vid_to_voltage)(family, model, vid, voltage);
}


/* Calculates vid from voltage */
int msr_voltage_to_vid(float voltage, int *vid)
{
	return (*ptr_msr_voltage_to_vid)(family, model, voltage, vid);
}


/* Get cpu status (vid and fid) */
int msr_get_status(int cpuid, int *vid, int *fid)
{
	char msrfile[32];

	if (libcpupowerd_get_msrfile(cpuid, msrfile, sizeof(msrfile)) != LIBCPUPOWERD_OK)
		return NOTOK;

	return (*ptr_msr_get_status)(family, model, msrfile, vid, fid);
}


/* Set cpu status( vid and fid) */
int msr_set_status(int cpuid, int vid, int fid)
{
	char msrfile[32];

	if (libcpupowerd_get_msrfile(cpuid, msrfile, sizeof(msrfile)) != LIBCPUPOWERD_OK)
		return NOTOK;

	return (*ptr_msr_set_status)(family, model, msrfile, vid, fid);
}




