/************************************************************************/
/* Copyright (C) 2008-2009 Markus Strobl <markus.strobl@gmx.at>         */
/*                                                                      */
/* cpupowerd is free software; you can redistribute it and/or modify it */
/* under the terms of the GNU General Public License as published by    */
/* the Free Software Foundation; either version 2 of the License.       */
/*                                                                      */
/* cpupowerd is distributed in the hope that it will be useful, but     */
/* WITHOUT ANY WARRANTY; without even the implied warranty of           */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */
/*                                                                      */
/* You should have received a copy of the GNU General Public License    */
/* along with GNU Emacs; see the file COPYING.  If not, write to        */
/* the Free Software Foundation, Inc., 59 Temple Place - Suite 330,     */
/* Boston, MA 02111-1307, USA.                                          */
/************************************************************************/
/* $Id: msr_amd.c 369 2009-01-01 13:05:07Z markus $                     */
/************************************************************************/


/* Includes */
#include <stdio.h>
#include <unistd.h>
#include <math.h>
#include <time.h>
#include <sys/time.h>

#include "cpupowerd.h"
#include "libcpupowerd.h"
#include "msr_amd.h"


/* Defines */
#define MSR_FIDVID_CTL 		0xC0010041	/* MSR control register*/
#define MSR_FIDVID_STATUS 	0xC0010042	/* MSR status register */
#define MSR_PENDING 		0x80000000  /* CPU pending flag position in MSR status register */

#define FIVE_BITS				0x1F	/* K8 befor F model */
#define SIX_BITS				0x3F	/* K8 F model and higher */


/* Prototypes */
int msr_amd_tini(void);
int msr_amd_vid_to_voltage(int family, int model, int vid, float *voltage);
int msr_amd_voltage_to_vid(int family, int model, float voltage, int *vid);
int msr_amd_fid_to_frequency(int family, int model, int fid, int *frequency);
int msr_amd_frequency_to_fid(int family, int model, int frequency, int *fid);
int msr_amd_get_status(int family, int model, char *msrfile, int *vid, int *fid);
int msr_amd_set_status(int family, int model, char *msrfile, int vid, int fid);
static int msr_amd_pending(char *msrfile);
static int msr_amd_get_model_parameter(int family, int model, int *vid_step, int *stabilization_time, int *bits);
static int msr_amd_verify_voltage(int family, int model, float voltage);
static int msr_amd_verify_vid(int family, int model, int vid);
static int msr_amd_verify_frequency(int family, int model, int frequency);
static int msr_amd_verify_fid(int family, int model, int fid);


/* Functions */
int msr_amd_init(int (**ptr_msr_vid_to_voltage) (int, int, int, float *),
				 int (**ptr_msr_voltage_to_vid) (int, int, float, int *),
				 int (**ptr_msr_fid_to_frequency) (int, int, int, int *),
				 int (**ptr_msr_frequency_to_fid) (int, int, int, int *),
				 int (**ptr_msr_get_status) (int, int, char *, int *, int *),
				 int (**ptr_msr_set_status) (int, int, char *, int, int),
				 int (**ptr_msr_tini) (void))
{
	*ptr_msr_vid_to_voltage = msr_amd_vid_to_voltage;
	*ptr_msr_voltage_to_vid = msr_amd_voltage_to_vid;
	*ptr_msr_fid_to_frequency = msr_amd_fid_to_frequency;
	*ptr_msr_frequency_to_fid = msr_amd_frequency_to_fid;
	*ptr_msr_get_status = msr_amd_get_status;
	*ptr_msr_set_status = msr_amd_set_status;
	*ptr_msr_tini = msr_amd_tini;

	return OK;
}


/* Clean up */
int msr_amd_tini(void)
{
	/* For future releases */
	return OK;
}


/* Calculates voltage from vid */
int msr_amd_vid_to_voltage(int family, int model, int vid, float *voltage)
{
	if (vid > 31)
		*voltage = 1.550 - ((31 * 0.025) + ((vid - 31) * 0.0125));
	else
		*voltage = 1.550 - (vid * 0.025);

	if (msr_amd_verify_voltage(family, model, *voltage) != OK)
		return NOTOK;

	return OK;
}


/* Calculates vid from voltage */
int msr_amd_voltage_to_vid(int family, int model, float voltage, int *vid)
{
	float vid_tmp;

	if (voltage < 0.7750)
		vid_tmp = 31 + (0.7750 - voltage) / 0.0125;
	else
		vid_tmp = (1.5500 - voltage) / 0.025;

	/* Round float to int */
	*vid = (int)(vid_tmp + 0.5);

	if (msr_amd_verify_vid(family, model, *vid) != OK)
		return NOTOK;

	return OK;
}


/* Calculates frequency from fid */
int msr_amd_fid_to_frequency(int family, int model, int fid, int *frequency)
{
	*frequency = (fid + 8) * 100;

	if (msr_amd_verify_frequency(family, model, *frequency) != OK)
		return NOTOK;

	return OK;
}


/* Calculates fid from frequency */
int msr_amd_frequency_to_fid(int family, int model, int frequency, int *fid)
{
	*fid = (frequency/100) - 8;

	if (msr_amd_verify_fid(family, model, *fid) != OK)
		return NOTOK;

	return OK;
}


/* Get cpu status (vid and fid) */
int msr_amd_get_status(int family, int model, char *msrfile, int *vid, int *fid)
{
	int bits;
	int vid_step;
	int stabilization_time;
	unsigned long low;
	unsigned long high;

	if (msr_amd_get_model_parameter(family, model, &vid_step, &stabilization_time, &bits) != OK)
		return NOTOK;

	if (libcpupowerd_read_msr(msrfile, MSR_FIDVID_STATUS, &low, &high) != LIBCPUPOWERD_OK)
		return NOTOK;
  	
	*vid = (int)(high & bits);
	*fid = (int)(low & SIX_BITS);

	return OK;
}


/* Set cpu status (vid, vid_step and fid) */
int msr_amd_set_status(int family, int model, char *msrfile, int vid, int fid)
{
	int bits;
	int vid_step;
	int vid_current;
	int fid_current;
	int stabilization_time;
	unsigned long long msrvalue;

	if (msr_amd_get_model_parameter(family, model, &vid_step, &stabilization_time, &bits) != OK)
		return NOTOK;

	if (msr_amd_get_status(family, model, msrfile, &vid_current, &fid_current) != OK)
		return NOTOK;

	/* Yet only VID change are supported, in future fid will also supported */
	if (fid != fid_current)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Fid changing currently not supported!\n");
		return NOTOK;
	}

	if (vid == vid_current)
		return OK;

	/* Check for bottom vid limit */
	switch (family)
	{
		case AMD_K8:
			if (model < 0x4 && vid > 30)	/* Model -E */
			{
				libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Your AMD-Model have 0.8000 voltage as bottom limit!\n");
				return NOTOK;
			}
			break;
		default:
			libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "AMD-Model <%d> not supported!\n", model);
			return NOTOK;
	}

	/* Check fid */
	if (msr_amd_verify_fid(family, model, fid) != OK)
		return NOTOK;

	/* Check vid */
	if (msr_amd_verify_vid(family, model, vid) != OK)
		return NOTOK;
	

	/* see AMD documentation 32559.pdf
	side 297: Figure 10. Phase 1: CPU Voltage Transition Flow
	side 302: 10.5.7.2.1 Changing the VID
	Note: Software must hold the FID constant when changing the VID.
	To change the processor voltage:
	1. Write the following values to FIDVID_CTL (MSR C001_0041h):
	   – NewVID field (bits 13–8) with the VID code associated with the target voltage
	   – NewFID field (bits 5–0) with the CurrFID value indicated in the FIDVID_STATUS MSR
	   – InitFidVid bit (bit 16) set to 1. Setting this bit initiates the VID change.
	   – Clear all other bits to 0.
	2. Loop on reading the FidVidPending bit (bit 31) of FIDVID_STATUS (MSR C001_0042h) until
	   the bit returns 0. The FidVidPending bit stays set to 1 until the new VID code has been driven to
	   the voltage regulator.
	*/

	while (vid_current != vid)
	{
		if (vid_current > vid)
		{
			vid_current -= vid_step;
			if (vid_current < vid)
				vid_current = vid;
		}
		else if (vid_current < vid)
		{
			vid_current += vid_step;
			if (vid_current > vid)
				vid_current = vid;
		}

		msrvalue = ((1ULL << 16) | ((vid_current) << 8) | fid);

		if (libcpupowerd_write_msr(msrfile, MSR_FIDVID_CTL, &msrvalue) != LIBCPUPOWERD_OK)
			return NOTOK;

		/* Look for pending */
		if (msr_amd_pending(msrfile) != OK)
			return NOTOK;


		/* Wait VoltageStabilizationTime in useconds */
		if (stabilization_time > 100)
			usleep(stabilization_time);
		else
			libcpupowerd_usleep(stabilization_time);
	}

	if (msr_amd_get_status(family, model, msrfile, &vid_current, &fid_current) != OK)
		return NOTOK;

	if (vid_current != vid)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Vid can't be set correctly. VID should be %d, but is %d!\n", vid, vid_current);
		return NOTOK;
	}

	if (fid_current != fid)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Fid can't be set correctly. FID should be %d, but is %d!\n", fid, fid_current);
		return NOTOK;
	}

	return OK;
}


/* Controls the cpu pending */
static int msr_amd_pending(char *msrfile)
{
	int i;
	unsigned long low;
	unsigned long high;

	i = 10000;

	do
	{
		if (libcpupowerd_read_msr(msrfile, MSR_FIDVID_STATUS, &low, &high) != LIBCPUPOWERD_OK)
			return NOTOK;
	}
	while (low & MSR_PENDING && --i);

	if (i <= 0)
	{
		libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Hardware error - pending bit very stuck!\n");
		return NOTOK;
	}
	
	return OK;
}


/* Get model parameter */
static int msr_amd_get_model_parameter(int family, int model, int *vid_step, int *stabilization_time, int *bits)
{
	int result;

	switch (family)
	{
		case AMD_K8:
			if (model < 0x4)	/* Model -E */
			{
				*stabilization_time = 100; /* in useconds */
				*bits = FIVE_BITS;
			}
			else				/* Model F- */
			{
				*stabilization_time = 40;  /* in useconds */
				*bits = SIX_BITS;
			}
			*vid_step = 1;
			result = OK;
			break;
		default:
			libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "AMD-Model <%d> not supported!\n", model);
			result = NOTOK;
	}

	return result;
}


/* Verify voltage */
static int msr_amd_verify_voltage(int family, int model, float voltage)
{
	int result;

	switch (family)
	{
		case AMD_K8:
			result = OK;
			if (model < 0x4) /* Model -E */
			{
				if (voltage > 1.550 || voltage < 0.8000)
				{
					libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Voltage (%.4f) out of range!\n", voltage);
					result = NOTOK;
				}
			}
			else /* Model F- */
			{
				if (voltage > 1.550 || voltage < 0.3750)
				{
					libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Voltage (%.4f) out of range!\n", voltage);
					result = NOTOK;
				}
			}
			
			break;
		default:
			libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Voltage check for AMD-Family <%d> not supported!\n", family);
			result = NOTOK;
	}

	return result;
}


/* Verify vid */
static int msr_amd_verify_vid(int family, int model, int vid)
{
	int result;

	switch (family)
	{
		case AMD_K8:
			result = OK;
			if (model < 0x4) /* Model -E */
			{
				if (vid > 30 || vid < 0)
				{
					libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "VID (%d) out of range!\n", vid);
					result = NOTOK;
				}
			}
			else /* Model F- */
			{
				if (vid > 63 || vid < 0)
				{
					libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "VID (%d) out of range!\n", vid);
					result = NOTOK;
				}
			}
			
			break;
		default:
			libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "VID check for AMD-Family <%d> not supported!\n", family);
			result = NOTOK;
	}

	return result;
}


/* Verify frequency */
static int msr_amd_verify_frequency(int family, int model, int frequency)
{
	int result;

	switch (family)
	{
		case AMD_K8:
			if (frequency < 800 || frequency > 5000)
			{
				libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Frequency (%d) out of range!\n", frequency);
				result = NOTOK;
			}
			else 
				result = OK;
			
			break;
		default:
			libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "Frequency check for AMD-Family <%d> not supported!\n", family);
			result = NOTOK;
	}

	return result;
}


/* Verify fid */
static int msr_amd_verify_fid(int family, int model, int fid)
{
	int result;

	switch (family)
	{
		case AMD_K8:
			if (fid < 0 || fid > 42)
			{
				libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "FID (%d) out of range!\n", fid);
				result = NOTOK;
			}
			else 
				result = OK;
			
			break;
		default:
			libcpupowerd_message(LIBCPUPOWERD_MESSAGE_ERROR, "FID check for AMD-Family <%d> not supported!\n", family);
			result = NOTOK;
	}

	return result;
}




