#! /bin/sh
##########################################################################
#  Copyright (C) 2008-2009 Markus Strobl <markus.strobl@gmx.at>          #
#                                                                        #
#  cpupowerd is free software; you can redistribute it and/or modify it  #
#  under the terms of the GNU General Public License as published by     #
#  the Free Software Foundation; either version 2 of the License.        #
#                                                                        #
#  cpupowerd is distributed in the hope that it will be useful, but      #
#  WITHOUT ANY WARRANTY; without even the implied warranty of            #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#  GNU General Public License for more details.                          #
#                                                                        #
#  You should have received a copy of the GNU General Public License     #
#  along with GNU Emacs; see the file COPYING.  If not, write to         #
#  the Free Software Foundation, Inc., 59 Temple Place - Suite 330,      #
#  Boston, MA 02111-1307, USA.                                           #
##########################################################################
#  $Id: cpupowerd.sh 369 2009-01-01 13:05:07Z markus $                   #
##########################################################################

CPUPOWERD_STARTPARAMETERS="-d -c /etc/cpupowerd.conf"

load_moduls()
{
	CPUINFO="/proc/cpuinfo"
	PATH_MODULS="/lib/modules/`uname -r`/kernel/drivers/cpufreq"

	# Load cpufreq kernel moduls if available
	if [ -d $PATH_MODULS ]; then		
		for modul in `ls $PATH_MODULS | grep -E '.ko$'`; do
			modul=${modul%.ko*}
			lsmod | cut -d ' ' -f 1 | grep -q -w $modul || modprobe $modul > /dev/null
		done
	fi;	

	# Load cpu vendor specific kernel modul
	VENDOR_ID=`grep -E '^vendor_id[^:]+:' "$CPUINFO" | head -1 | sed -e 's/^.*: //;'`
	MODEL_NAME=`grep '^model name' "$CPUINFO" | head -1 | sed -e 's/^.*: //;'`
	CPU_FAMILY=`grep '^cpu family' /proc/cpuinfo | head -1 | sed -e 's/^.*: //;'`

	case $VENDOR_ID in
		AuthenticAMD*)
			case $CPU_FAMILY in
				15) # K8
					modul="powernow-k8"
					;;
				*)
					echo "AMD family $CPU_FAMILY not supported!"
					return 1
					;;
			esac
		;;
		*)
			echo "$VENDOR_ID not supported!"
			return 1
		;;
	esac
	lsmod | cut -d ' ' -f 1 | grep -q $modul || modprobe $modul > /dev/null 2>&1

	# Load acpi kernel modul, if governor isn't available
	if [ ! -f /sys/devices/system/cpu/cpu0/cpufreq/scaling_governor ]; then
		modprobe acpi > /dev/null 2>&1 || true
	fi

	# Load msr kernel modul, if modul isn't loaded
	lsmod | cut -d ' ' -f 1 | grep -q msr || modprobe msr > /dev/null 2>&1

	return 0	
}

is_kernel_interfaces()
{
	PATH_CPUFREQ="/sys/devices/system/cpu/cpu0/cpufreq"
	PATH_MSR="/dev/cpu/0"

	is_file ${PATH_CPUFREQ}/scaling_available_frequencies
	if [ $? -ne 1 ]; then
		echo "${PATH_CPUFREQ}/scaling_available_frequencies not available!"
		return 0
	fi

	is_file ${PATH_CPUFREQ}/scaling_available_governors
	if [ $? -ne 1 ]; then
		echo "${PATH_CPUFREQ}/scaling_available_governors not available!"
		return 0
	fi

	is_file ${PATH_CPUFREQ}/scaling_cur_freq
	if [ $? -ne 1 ]; then
		echo "${PATH_CPUFREQ}/scaling_cur_freq not available!"
		return 0
	fi

	is_file ${PATH_CPUFREQ}/scaling_governor
	if [ $? -ne 1 ]; then
		echo "${PATH_CPUFREQ}/scaling_governor not available!"
		return 0
	fi

	is_file ${PATH_CPUFREQ}/affected_cpus
	if [ $? -ne 1 ]; then
		echo "${PATH_CPUFREQ}/affected_cpus not available!"
		return 0
	fi

	is_file ${PATH_MSR}/msr
	if [ $? -ne 1 ]; then
		echo "${PATH_MSR}/msr not available!"
		return 0
	fi

	cat ${PATH_CPUFREQ}/scaling_available_governors | grep -q userspace > /dev/null 2>&1
	if [ $? -ne 0 ]; then
		echo "Governor userspace isn't available in ${PATH_CPUFREQ}/scaling_available_governors!"
		return 0
	fi

	return 1
}

is_file()
{
	if [ -z "$1" ]; then
		return 0
	fi

	# kernel modul load take time
	for a in 1 2 3 4 5 6 7 8 9 10
	do
		if [ -e "$1" ]; then
			return 1
		fi

		sleep 1
	done

	return 0
}

is_environment()
{
	which powernowd > /dev/null 2>&1
	if [ $? -eq 0 ]; then
		pidof `which powernowd` > /dev/null 2>&1
		if [ $? -eq 0 ]; then
			echo "Please stop powernowd before you start cpupowerd!"
			return 0
		fi
	fi

	which cpufreqd > /dev/null 2>&1
	if [ $? -eq 0 ]; then
		pidof `which cpufreqd` > /dev/null 2>&1
		if [ $? -eq 0 ]; then
			echo "Please stop cpufreqd before you start cpupowerd!"
			return 0
		fi
	fi

	which cpudyn > /dev/null 2>&1
	if [ $? -eq 0 ]; then
		pidof `which cpudyn` > /dev/null 2>&1
		if [ $? -eq 0 ]; then
			echo "Please stop cpudyn before you start cpupowerd!"
			return 0
		fi
	fi
	
	which cpupowerd > /dev/null 2>&1
	if [ $? -eq 0 ]; then
		pidof `which cpupowerd` > /dev/null 2>&1
		if [ $? -eq 0 ]; then
			echo "Please stop cpupowerd before you start cpupowerd again!"
			return 0
		fi
	fi

	return 1
}

PATH=/bin:/sbin:/usr/bin:/usr/sbin:/usr/local/bin:/usr/local/sbin

case "$1" in
	start)
		# Check environment
		# sleep 30
		# killall hald-addon-cpufreq
		is_environment
		if [ $? -ne 1 ]; then
			echo "Environment for cpupowerd isn't ok!"
			exit 1
		fi

		# Load kernel moduls
		load_moduls
		if [ $? -ne 0 ]; then
			echo "Couldn't load kernel moduls!"
			exit 1
		fi

		# Check kernel interfaces
		is_kernel_interfaces
		if [ $? -ne 1 ]; then
			echo "Some required kernel interfaces not available!"
			exit 1
		fi

		# start cpupowerd
		`which cpupowerd` $CPUPOWERD_STARTPARAMETERS
		echo "cpupowerd $CPUPOWERD_STARTPARAMETERS successfully started."
	;;
	stop)
		# stop all started cpupowerd
		for cpupowerdpid in $(pidof `which cpupowerd`)
		do
			kill $cpupowerdpid > /dev/null 2>&1
			if [ $? -ne 0 ]; then
				echo "Stopping cpupowerd with pid $cpupowerdpid failed!"
			else
				echo "cpupowerd with pid $cpupowerdpid successfully stoped."
			fi
		done
	;;
	restart)
		$0 stop
		sleep 1
		$0 start
	;;
	*)
		echo "Usage: $0 {start|stop|restart}" >&2
		exit 3
	;;
esac



